Require Import FJ_tactics.
Require Import List.
Require Import Functors.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import MonadLib.
Require Import Arith.
Require Import Exception.

Section Exception_Arith.

  Variable D : Set -> Set.
  Context {Fun_D : Functor D}.
  Context {Sub_AType_D : AType :<: D}.

  Variable F : Set -> Set.
  Context {Fun_F : Functor F}.
  Context {Sub_ExceptE_F : ExceptE D :<: F}.
  Context {Sub_Arith_F : Arith :<: F}.

  (* Constructor + Universal Property. *)
   Context {WF_Sub_ExceptE_F : WF_Functor _ _ Sub_ExceptE_F _ _}.
   Context {WF_Sub_Arith_F : WF_Functor _ _ Sub_Arith_F _ _}.

   Variable V : Set -> Set.
   Context {Fun_V : Functor V}.
   Context {Sub_StuckValue_V : StuckValue :<: V}.
   Context {Sub_NatValue_V : NatValue :<: V}.

  Variable MT : Set -> Set.
  Context `{Fail_MT : FailMonad MT}.

  Context {Inj_MT : InjMonad MT}.
  Context {Reasonable_MT : Reasonable_Monad MT}.

  Context {eq_DType_DT : forall T, FAlgebra eq_DTypeName T (eq_DTypeR D) D}.
  Context {eq_DType_eq_DT : PAlgebra eq_DType_eqName D D (UP'_P (eq_DType_eq_P D))}.
  Context {eq_DType_neq_D : PAlgebra eq_DType_neqName D D (UP'_P (eq_DType_neq_P D))}.

  Variable ME : Set -> Set.
  Context `{Exception_ME : Exception ME unit}.
  Context {Fail_ME : FailMonad ME}.

  Context {ME_eq_dec : forall (A : Set) (mte : ME A),
    (exists a, mte = return_ a) \/ (mte = fail) \/ (mte = throw tt)}.
  Context {Reasonable_ME : Reasonable_Monad ME}.
  Context {fail_neq_throw : forall A, fail (M := ME) <> throw (A := A) tt}.

  Context {evalM_F : FAlgebra EvalName (Exp F) (evalMR V ME) F}.
  Context {WF_eval_ExceptE_F : @WF_FAlgebra EvalName _ _ (ExceptE D) F
    Sub_ExceptE_F (MAlgebra_evalM_ExceptE D V ME _) evalM_F}.
  Context {WF_eval_Arith_F : @WF_FAlgebra EvalName _ _ Arith F
    Sub_Arith_F (MAlgebra_evalM_Arith V ME _) evalM_F}.

  Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR D MT) F}.
  Context {WF_ExceptE_Typeof_F : forall T, @WF_FAlgebra TypeofName _ _ (ExceptE D) F
    Sub_ExceptE_F (MAlgebra_typeof_ExceptE _ MT _) (Typeof_F T)}.
  Context {WF_Arith_Typeof_F : forall T, @WF_FAlgebra TypeofName _ _ Arith F
    Sub_Arith_F (MAlgebra_typeof_Arith _ MT _) (Typeof_F T)}.

    Context {evalM_F' : forall T, FAlgebra EvalName T (evalMR V ME) F}.
    Context {WF_ExceptE_eval_F' : forall T, @WF_FAlgebra EvalName _ _ (ExceptE D) F
      Sub_ExceptE_F (MAlgebra_evalM_ExceptE D V ME _) (evalM_F' T)}.
    Context {WF_Arith_eval_F' : forall T, @WF_FAlgebra EvalName _ _ Arith F
      Sub_Arith_F (MAlgebra_evalM_Arith V ME _) (evalM_F' T)}.

    Variable TypContext : Set.
    
    Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
    Context {funWFV : iFunctor WFV}.

    Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> WFValueM_i D V MT ME TypContext -> Prop.
    Context {funWFVM : iFunctor WFVM}.
    
    Context {TypContextCE : ConsExtensionC TypContext}.

    Context {Sub_WFVM_Base_WFVM : Sub_iFunctor (WFValueM_base D V MT ME _ WFV) WFVM}.
    Context {Sub_WFVM_Except_WFVM : Sub_iFunctor (WFValueM_Except D V MT ME _) WFVM}.
    
  Lemma functional_extensionality2 : forall (A B C : Set) (f g : A -> B -> C),
    (forall a b, f a b = g a b) -> f = g.
  Proof.
    repeat (intros; apply functional_extensionality_dep); auto.
  Qed.

  Lemma rewrite_do2 : forall (M : Set -> Set)
    {Fun_M : Functor M}
    {Monad_M : Monad M},
    forall (A B C : Set) (n : M A) (o : M B) (m m' : A -> B -> M C),
      m = m' -> (do n' <- n; do o' <- o; m n' o') = (do n' <- n; do o' <- o; m' n' o').
    intros; rewrite H; reflexivity.
  Qed.

  Global Instance WFVM_Except_eval_Soundness_Addm : iPAlgebra Arith_eval_Soundness_Addm_Name
    (Arith_eval_Soundness_Addm_P V ME D MT TypContext WFVM) (WFValueM_Except D V MT ME TypContext).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
      try eassumption; unfold Arith_eval_Soundness_Addm_P; split; simpl; intros.
    (* throw case *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
    rewrite bind_throw.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
    (* catch case *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor; eauto.
    apply (proj1 H0).
    intros t Sigma' Sub_Sig_Sig'; apply (proj1 (H1 t Sigma' Sub_Sig_Sig')).
    repeat rewrite <- associativity.
    rewrite rewrite_do with (n := kT')
      (m' := fun x =>
        do y <- (do te <- mte;
          do th <- mth;
            (if eq_DType D (proj1_sig te) th then return_ te else fail));
        do tm <- kT x y;
        do tn <- n2;
          (if isTNat D (proj1_sig tm)
            then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
            else fail));
      [idtac | apply functional_extensionality; intros; rewrite <- associativity; auto].
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2.
    intros; apply f_equal.
    repeat rewrite associativity,
      (eq_bind_fmap _ _ _ _ (do U <- kT'; kT U _) _
      (fun tm => do tn <- n2;
        (if isTNat D tm
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)) (@proj1_sig _ _)); auto.
    unfold Names.DType, UP'_F in *|-*; rewrite (kt_eq _ _ H3); auto.
    repeat rewrite associativity.
    assert (do tm <- do U <- kT'; mte >>= kT U;
     do tn <- n2;
     (if isTNat D (proj1_sig tm)
      then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
      else fail) = do U <- kT';
              do y <- mte;
              do tm <- kT U y;
              do tn <- n2;
              (if isTNat D (proj1_sig tm)
               then
                if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
               else fail)) as mt_eq by
    (rewrite <- associativity; apply f_equal; apply functional_extensionality;
      intros; rewrite <- associativity; auto).
    unfold DType in *|-*; rewrite <- mt_eq; apply H0; eauto.
    intros; repeat rewrite associativity.
    assert (do U <- kT';
               do y <- mth;
               do tm <- kT U y;
               do tn <- n2;
               (if isTNat D (proj1_sig tm)
                then
                 if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
                else fail) =
               do tm <- do U <- kT'; mth >>= kT U;
                 do tn <- n2;
                   (if isTNat D (proj1_sig tm)
                     then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
                     else fail)) as mth_eq by
    (rewrite <- associativity; apply f_equal; apply functional_extensionality;
      intros; rewrite <- associativity; auto).
    unfold DType in *|-*; rewrite mth_eq; apply H1; eauto.
    intros; eapply H2; eapply ConsExtension_trans; eauto.
  Qed.

  Context {WFV_Weaken_WFV : iPAlgebra WFValue_Weaken_Name (WFValue_Weaken_P D V _ WFV) WFV}.

  Global Instance WFVM_Except_eval_Soundness_Addn :
    iPAlgebra Arith_eval_Soundness_Addn_Name
    (Arith_eval_Soundness_Addn_P V ME D MT TypContext WFV WFVM) (WFValueM_Except D V MT ME TypContext).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
      try eassumption; unfold Arith_eval_Soundness_Addn_P; split; simpl; intros.
    (* throw case *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
    rewrite bind_throw.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
    (* catch case *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor; eauto.
    apply (proj1 H0).
    intros t Sigma' ConsExt_Sig_Sig'; apply (proj1 (H1 t Sigma' ConsExt_Sig_Sig')).
    repeat rewrite <- associativity.
    assert (do tm <- fmap (@proj1_sig _ _) mt';
      do tn <-
        do U <- kT';
          (do te <- mte;
           do th <- mth;
              (if eq_DType D (proj1_sig te) th then return_ te else fail)) >>= kT U;
          (if isTNat D (proj1_sig tm)
            then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
            else fail) =
          do tm <- (do tm' <- fmap (@proj1_sig _ _) mt';
            do U <- kT'; return_ (tm', U));
          (do te <- mte; do th <- mth;
            (if eq_DType D (proj1_sig te) th then return_ te else fail)) >>=
          fun u => (do tn <- kT (snd tm) u; (if isTNat D (proj1_sig (fst tm))
              then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
              else fail))) as mt_eq by
    (rewrite <- associativity; apply f_equal; apply functional_extensionality; intros;
      repeat rewrite <- associativity; apply f_equal; apply functional_extensionality; intros;
        rewrite <- left_unit, associativity; simpl; auto); rewrite mt_eq; clear mt_eq.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor 2; eauto.
    intros; repeat rewrite <- associativity; apply f_equal.
    apply f_equal; apply functional_extensionality; intros.
    repeat rewrite <- associativity.
    rewrite rewrite_do with (n := kT')
      (m' := fun x0 => do tn <- kT x0 T0;
        (if isTNat D (proj1_sig x)
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)).
    rewrite rewrite_do with (n := kT') (m := fun x0 => do U <- return_ (x, x0); _)
      (m' := fun x0 => do tn <- kT x0 T';
        (if isTNat D (proj1_sig x)
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)).
    repeat rewrite associativity,
      (eq_bind_fmap _ _ _ _ (do U <- kT'; kT U _) _
        (fun tn => if isTNat D (proj1_sig x)
          then if isTNat D tn then return_ (tnat' D) else fail
          else fail) (@proj1_sig _ _)); auto.
    unfold Names.DType, UP'_F in *|-*; rewrite (kt_eq _ _ H2); auto.
    apply functional_extensionality; intros; rewrite <- left_unit; reflexivity.
    apply functional_extensionality; intros; rewrite <- left_unit; reflexivity.
    repeat rewrite associativity.
    assert (do tm <- fmap (@proj1_sig _ _) mt';
      do tn <- do U <- kT'; mte >>= kT U;
        (if isTNat D (proj1_sig tm)
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail) = do U <-
               do tm' <- fmap (@proj1_sig _ _) mt'; do U <- kT'; return_ (tm', U);
               do u <- mte;
               do tn <- kT (snd U) u;
               (if isTNat D (proj1_sig (fst U))
                then
                 if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
                else fail)) as mt_eq by
    (rewrite <- associativity; apply f_equal; apply functional_extensionality;
      intros; rewrite <- associativity; auto;
        rewrite <- associativity; apply f_equal; apply functional_extensionality;
          intros; rewrite <- associativity; auto; rewrite <- left_unit; simpl; auto).
    unfold DType in *|-*; rewrite <- mt_eq; apply H0; eauto.
    intros; rewrite associativity.
    assert (do tm <- fmap (@proj1_sig _ _) mt';
      do tn <- do U <- kT'; mth >>= kT U;
        (if isTNat D (proj1_sig tm)
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail) = do U <-
               do tm' <- fmap (@proj1_sig _ _) mt'; do U <- kT'; return_ (tm', U);
               do u <- mth;
               do tn <- kT (snd U) u;
               (if isTNat D (proj1_sig (fst U))
                then
                 if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
                else fail)) as mt_eq by
    (rewrite <- associativity; apply f_equal; apply functional_extensionality;
      intros; rewrite <- associativity; auto;
        rewrite <- associativity; apply f_equal; apply functional_extensionality;
          intros; rewrite <- associativity; auto; rewrite <- left_unit; simpl; auto).
    unfold DType in *|-*; rewrite <- mt_eq; eapply H1; eauto.
    apply (WFV_Weaken _ _ _ WFV _ _ WF_v_T); eauto.
  Qed.

    (* Doesn't work...
       Global Instance WFVM_Except_eval_Soundness_Addn' :
    iPAlgebra Arith_eval_Soundness_Addn'_Name
    (Arith_eval_Soundness_Addn'_P V ME D MT WFV WFVM) (WFValueM_Except D V MT ME).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
      try assumption; unfold Arith_eval_Soundness_Addn'_P; simpl; intros.
    (* throw case *)
    rewrite bind_throw.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
    (* catch case *)
    repeat rewrite <- associativity.
    assert (do tm <- fmap (@proj1_sig _ _) mt';
      do tn <- do x <- mte; (do th <- mth;
        (if eq_DType D (proj1_sig x) th then return_ x else fail)) >>= kT;
      (if isTNat D (proj1_sig tm)
        then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
        else fail) =
      (do tm <- fmap (@proj1_sig _ _) mt';
        do tn <- mte;
          (do th <- mth; (if eq_DType D (proj1_sig tn) th then return_ tn else fail)) >>=
          fun x => do tn <- kT x;
            (if isTNat D (proj1_sig x)
              then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
              else fail))).
    rewrite (eq_bind_fmap _ _ _ _ mt' _
      (fun tm => do tn <- typeof (Typeof_E := Typeof_F) D F MT n;
        (if isTNat D tm
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)) (@proj1_sig _ _)); auto.
    apply f_equal; apply functional_extensionality; intros;
      repeat (rewrite <- associativity; apply f_equal;
        apply functional_extensionality; intros); auto.
    apply f_equal; apply functional_extensionality; intros.
    apply f_equal.

 as kT_eq by
    (apply f_equal; apply functional_extensionality; intros;
      (repeat (rewrite <- associativity; apply f_equal;
        apply functional_extensionality; intros); auto)); rewrite kT_eq; clear kT_eq.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
    intros; apply f_equal.
    rewrite (eq_bind_fmap _ _ _ _ (kT T) _
      (fun tm => do tn <- typeof (Typeof_E := Typeof_F) D F MT n;
        (if isTNat D tm
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)) (@proj1_sig _ _)); auto.
    rewrite (eq_bind_fmap _ _ _ _ (kT T') _
      (fun tm => do tn <- typeof (Typeof_E := Typeof_F) D F MT n;
        (if isTNat D tm
          then if isTNat D (proj1_sig tn) then return_ (tnat' D) else fail
          else fail)) (@proj1_sig _ _)); auto.
    unfold Names.DType, UP'_F; rewrite (kt_eq _ _ H3); auto.
    repeat rewrite associativity; eapply H0; eauto.
    intros; repeat rewrite associativity; eapply H1; eauto.
  Qed. *)

  End Exception_Arith.

(* Hint Extern 0 =>
  intros; match goal with
            |- (PAlgebra eval_Soundness_alg_Name
              (sig (UP'_P2 (eval_alg_Soundness_P _ _ _ _ _ _ _ _ _ _ _ _ _ _))) Arith) =>
            eapply Arith_eval_Soundness_alg; eauto with typeclass_instances
          end : typeclass_instances. *)


(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
