Require Import Ascii.
Require Import String.
Require Import FJ_tactics.
Require Import List.
Require Import Functors.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import MonadLib.

Section Ref.

  (* ============================================== *)
  (* TYPES                                          *)
  (* ============================================== *)

  (* The Reference Type. *)
    Inductive RefType (A : Set) : Set :=
    | TRef : A -> RefType A.

    Definition RefType_fmap (A B : Set) (f : A -> B) (rta : RefType A)
      : RefType B :=
      match rta with
        | TRef a => TRef B (f a)
      end.

    Global Instance RefType_Functor : Functor RefType :=
      {| fmap := RefType_fmap |}.
    Proof.
      destruct a; reflexivity.
    (* fmap id *)
      destruct a; reflexivity.
    Defined.

    Variable D : Set -> Set.
    Context {Fun_D : Functor D}.
    Definition DType := DType D.
    Context {Sub_RefType_D  : RefType :<: D}.

    (* Constructor + Universal Property. *)
    Context {WF_Sub_RefType_D : WF_Functor _ _ Sub_RefType_D _ _}.

    Definition tref' (t : DType) : DType := inject' (TRef _ t).
    Definition tref (t : Fix D)
      {t_UP' : Universal_Property'_fold t}
      : Fix D := proj1_sig (tref' (exist _ _ t_UP')).
    Global Instance UP'_tref
      (t : Fix D)
      {t_UP' : Universal_Property'_fold t}
      : Universal_Property'_fold (tref t) := proj2_sig (tref' (exist _ _ t_UP')).

  (* Induction Principle for Ref Types. *)
    Definition ind_alg_RefType
      (P : forall d : Fix D, Universal_Property'_fold d -> Prop)
      (H : forall t (IHt : UP'_P P t),
        UP'_P P (@tref _ (proj1_sig IHt)))
      (d : RefType (sig (UP'_P P))) : sig (UP'_P P) :=
      match d with
        | TRef t => exist (UP'_P P) (tref (proj1_sig t)) (H (proj1_sig t) (proj2_sig t))
      end.

    Lemma ind_PAlg_RefType 
      {Sub_RefType_D' : RefType :<: D}
      (eq_Sub_RefType : forall A (d : RefType A),
        inj (Sub_Functor := Sub_RefType_D) d = inj (Sub_Functor := Sub_RefType_D') d)
      (Name : Set)
      (P : forall e : Fix D, Universal_Property'_fold e -> Prop)
      (H : forall t (IHt : UP'_P P t),
        UP'_P P (@tref _ (proj1_sig IHt)))
      :
      PAlgebra (sub_F_G := Sub_RefType_D') Name RefType D (UP'_P P).
      econstructor 1 with (p_algebra := ind_alg_RefType P H).
      destruct e; simpl; unfold tref, tref'; simpl;
        rewrite wf_functor; simpl; rewrite eq_Sub_RefType; reflexivity.
    Qed.

  (* Type Equality Section. *)
    Definition isTRef : Fix D -> option DType :=
      fun typ =>
        match project typ with
          | Some (TRef t) => Some t
          | None      => None
        end.
    
    Definition RefType_eq_DType  (R : Set) (rec : R -> eq_DTypeR D)
      (e : RefType R) : eq_DTypeR D :=
      match e with
        | TRef t => fun t' =>
          match (isTRef (proj1_sig t')) with
            | Some t'' => rec t t''
            | None => false
          end
      end.
    
    Global Instance MAlgebra_eq_DType_TRef T:
      FAlgebra eq_DTypeName T (eq_DTypeR D) RefType :=
      {| f_algebra := RefType_eq_DType T|}.
    
    Context {eq_DType_DT : forall T, FAlgebra eq_DTypeName T (eq_DTypeR D) D}.
    Context {WF_DType_eq_DT : forall T, @WF_FAlgebra eq_DTypeName T _ _ _
      Sub_RefType_D (MAlgebra_eq_DType_TRef T) (eq_DType_DT _)}.
    
    Global Instance PAlgebra_eq_DType_eq_RefType
      {Sub_RefType_D' : RefType :<: D}
      (eq_Sub_RefType : forall A (d : RefType A),
        inj (Sub_Functor := Sub_RefType_D) d = inj (Sub_Functor := Sub_RefType_D') d) :
      PAlgebra (sub_F_G := Sub_RefType_D') eq_DType_eqName RefType D (UP'_P (eq_DType_eq_P D)).
    Proof.
      apply ind_PAlg_RefType; eauto.
      unfold UP'_P; econstructor.
      unfold eq_DType_eq_P; intros.
      unfold eq_DType, mfold, tref, tref', inject' in H; simpl in H;
        repeat rewrite wf_functor in H; simpl in H; unfold in_t in H.
      rewrite (wf_algebra (WF_FAlgebra := WF_DType_eq_DT _)) in H; simpl in H.
      unfold isTRef in H.
      caseEq (project (subGF := Sub_RefType_D) (proj1_sig d2)); rewrite H0 in H;
        try discriminate; destruct r.
      apply project_inject in H0.
      destruct IHt as [t_UP' IHt]; unfold eq_DType_eq_P in IHt.
      rewrite H0; unfold tref, tref', inject; simpl;
        repeat rewrite wf_functor; simpl; unfold RefType_fmap; simpl.
      rewrite (IHt s H); auto.
      auto with typeclass_instances.
      exact (proj2_sig d2).
    Qed.
    
    Global Instance PAlgebra_eq_DType_neq_RefType 
      {Sub_RefType_D' : RefType :<: D}
      (eq_Sub_RefType : forall A (d : RefType A),
        inj (Sub_Functor := Sub_RefType_D) d = inj (Sub_Functor := Sub_RefType_D') d) :
      PAlgebra (sub_F_G := Sub_RefType_D') eq_DType_neqName RefType D (UP'_P (eq_DType_neq_P D)).
    Proof.
      apply ind_PAlg_RefType; eauto.
      unfold UP'_P; econstructor.
      unfold eq_DType_neq_P; intros.
      unfold eq_DType, mfold, tref, tref', inject' in H; simpl in H;
        repeat rewrite wf_functor in H; simpl in H; unfold in_t in H.
      rewrite (wf_algebra (WF_FAlgebra := WF_DType_eq_DT _)) in H; simpl in H.
      destruct IHt as [t_UP' IHt]; unfold eq_DType_neq_P in IHt.
      unfold not; intros; rewrite <- H0 in H.
      unfold isTRef, project, tref, inject in H; simpl in H.
      rewrite out_in_fmap in H; repeat rewrite wf_functor in H;
        simpl in H; rewrite prj_inj in H.
      unfold eq_DType in IHt; apply (IHt _ H).
      rewrite in_out_UP'_inverse; eauto.
    Qed.
        
    (* The Unit Type. *)
    Inductive UnitType (A : Set) : Set :=
    | TUnit : UnitType A.

    Definition UnitType_fmap (A B : Set) (f : A -> B) (rta : UnitType A)
      : UnitType B :=
      match rta with
        | TUnit => TUnit B
      end.

    Global Instance UnitType_Functor : Functor UnitType :=
      {| fmap := UnitType_fmap |}.
    Proof.
      destruct a; reflexivity.
    (* fmap id *)
      destruct a; reflexivity.
    Defined.

    Context {Sub_UnitType_D  : UnitType :<: D}.

    (* Constructor + Universal Property. *)
    Context {WF_Sub_UnitType_D : WF_Functor _ _ Sub_UnitType_D _ _}.

    Definition tunit' : DType := inject' (TUnit _).
    Definition tunit : Fix D := proj1_sig tunit'.
    Global Instance UP'_tunit : Universal_Property'_fold tunit :=
      proj2_sig tunit'.

  (* Induction Principle for Unit Types. *)
    Definition ind_alg_UnitType
      (P : forall d : Fix D, Universal_Property'_fold d -> Prop)
      (H : UP'_P P tunit)
      (d : UnitType (sig (UP'_P P))) : sig (UP'_P P) :=
      match d with
        | TUnit => exist (UP'_P P) tunit H
      end.

    Lemma ind_PAlg_UnitType
      {Sub_UnitType_D' : UnitType :<: D}
      (eq_Sub_UnitType : forall A (d : UnitType A),
        inj (Sub_Functor := Sub_UnitType_D) d = inj (Sub_Functor := Sub_UnitType_D') d)
      (Name : Set)
      (P : forall e : Fix D, Universal_Property'_fold e -> Prop)
      (H : UP'_P P tunit)
      :
      PAlgebra (sub_F_G := Sub_UnitType_D') Name UnitType D (UP'_P P).
      econstructor 1 with (p_algebra := ind_alg_UnitType P H).
      destruct e; simpl; unfold tunit, tunit'; simpl;
        rewrite wf_functor; simpl; rewrite eq_Sub_UnitType; auto.
    Qed.

  (* Type Equality Section. *)
  Definition isTUnit : Fix D -> bool :=
    fun typ =>
      match project typ with
        | Some TUnit => true
        | None           => false
      end.

  Definition UnitType_eq_DType  (R : Set) (rec : R -> eq_DTypeR D)
    (e : UnitType R) : eq_DTypeR D :=
    match e with
      | TUnit => fun t => isTUnit (proj1_sig t)
    end.

  Global Instance MAlgebra_eq_DType_TUnit T:
    FAlgebra eq_DTypeName T (eq_DTypeR D) UnitType :=
    {| f_algebra := UnitType_eq_DType T|}.

  Context {WF_Unit_eq_DT : forall T, @WF_FAlgebra eq_DTypeName T _ _ _
    Sub_UnitType_D (MAlgebra_eq_DType_TUnit T) (eq_DType_DT _)}.

  Global Instance PAlgebra_eq_DType_eq_UnitType 
    {Sub_UnitType_D' : UnitType :<: D}
    (eq_Sub_UnitType : forall A (d : UnitType A),
      inj (Sub_Functor := Sub_UnitType_D) d = inj (Sub_Functor := Sub_UnitType_D') d):
    PAlgebra (sub_F_G := Sub_UnitType_D') eq_DType_eqName UnitType D (UP'_P (eq_DType_eq_P D)).
  Proof.
    apply ind_PAlg_UnitType; eauto.
    unfold UP'_P; econstructor.
    unfold eq_DType_eq_P; intros.
    unfold eq_DType, mfold, tunit, tunit', inject' in H; simpl in H;
      repeat rewrite wf_functor in H; simpl in H; unfold in_t in H.
    rewrite (wf_algebra (WF_FAlgebra := WF_Unit_eq_DT _)) in H; simpl in H.
    unfold isTUnit in H.
    caseEq (project (subGF := Sub_UnitType_D) (proj1_sig d2)); rewrite H0 in H;
      try discriminate; destruct u.
    apply project_inject in H0.
    rewrite H0; unfold tunit, tunit', inject; simpl;
      repeat rewrite wf_functor; simpl; unfold UnitType_fmap; auto.
    auto with typeclass_instances.
    exact (proj2_sig d2).
  Qed.

  Global Instance PAlgebra_eq_DType_neq_UnitType 
    {Sub_UnitType_D' : UnitType :<: D}
    (eq_Sub_UnitType : forall A (d : UnitType A),
      inj (Sub_Functor := Sub_UnitType_D) d = inj (Sub_Functor := Sub_UnitType_D') d) :
    PAlgebra (sub_F_G := Sub_UnitType_D') eq_DType_neqName UnitType D (UP'_P (eq_DType_neq_P D)).
  Proof.
    apply ind_PAlg_UnitType; eauto.
    unfold UP'_P; econstructor.
    unfold eq_DType_neq_P; intros.
    unfold eq_DType, mfold, tunit, tunit', inject' in H; simpl in H;
      repeat rewrite wf_functor in H; simpl in H; unfold in_t in H.
    rewrite (wf_algebra (WF_FAlgebra := WF_Unit_eq_DT _)) in H; simpl in H.
    unfold not; intros; rewrite <- H0 in H.
    unfold isTUnit, project, tunit, inject in H; simpl in H.
    rewrite out_in_fmap in H; repeat rewrite wf_functor in H;
      simpl in H; rewrite prj_inj in H; discriminate.
  Qed.

  (* ============================================== *)
  (* EXPRESSIONS                                    *)
  (* ============================================== *)

  Inductive RefE (a : Set) : Set :=
  | Ref : a -> RefE a
  | DeRef : a -> RefE a
  | Assign : a -> a -> RefE a.

  Definition RefE_fmap (B C : Set) (f : B -> C) (Aa : RefE B) : RefE C :=
    match Aa with
      | Ref e => Ref _ (f e)
      | DeRef e => DeRef _ (f e)
      | Assign e1 e2 => Assign _ (f e1) (f e2)
    end.

  Global Instance RefE_Functor : Functor RefE :=
    {| fmap := RefE_fmap |}.
    destruct a; reflexivity.
    (* fmap id *)
    destruct a; reflexivity.
  Defined.

  Variable E : Set -> Set.
  Context {Fun_E : Functor E}.
  Definition Exp := Exp E.
  Context {Sub_RefE_E : RefE :<: E}.

  (* Constructor + Universal Property. *)
   Context {WF_Sub_RefE_E : WF_Functor _ _ Sub_RefE_E _ _}.

  Definition assign' (e1 e2 : Exp) : Exp :=  inject' (Assign _ e1 e2).

  Definition assign (e1 e2 : Fix E)
    {UP'_e1 : Universal_Property'_fold e1}
    {UP'_e2 : Universal_Property'_fold e2}
    : Fix E := proj1_sig (assign' (exist _ _ UP'_e1) (exist _ _ UP'_e2)).

   Global Instance UP'_assign {e1 e2 : Fix E}
     {UP'_e1 : Universal_Property'_fold e1}
     {UP'_e2 : Universal_Property'_fold e2}
     :
     Universal_Property'_fold (assign e1 e2) :=
     proj2_sig (assign' (exist _ _ UP'_e1) (exist _ _ UP'_e2)).

  Definition ref' (e : Exp) : Exp :=  inject' (Ref _ e).

  Definition ref (e : Fix E)
    {UP'_e : Universal_Property'_fold e}
    : Fix E := proj1_sig (ref' (exist _ _ UP'_e)).

   Global Instance UP'_ref {e : Fix E}
     {UP'_e : Universal_Property'_fold e}
     :
     Universal_Property'_fold (ref e) :=
     proj2_sig (ref' (exist _ _ UP'_e)).

  Definition deref' (e : Exp) : Exp :=  inject' (DeRef _ e).

  Definition deref (e : Fix E)
    {UP'_e : Universal_Property'_fold e}
    : Fix E := proj1_sig (deref' (exist _ _ UP'_e)).

   Global Instance UP'_deref {e : Fix E}
     {UP'_e : Universal_Property'_fold e}
     :
     Universal_Property'_fold (deref e) :=
     proj2_sig (deref' (exist _ _ UP'_e)).

  (* Induction Principles for RefE. *)
  Definition ind_alg_RefE
    (P : forall e : Fix E, Universal_Property'_fold e -> Prop)
    (H : forall e
      (IHe : UP'_P P e),
      UP'_P P (@ref e (proj1_sig IHe)))
    (H0 : forall e
      (IHe : UP'_P P e),
      UP'_P P (@deref e (proj1_sig IHe)))
    (H1 : forall e1 e2
      (IHe1 : UP'_P P e1)
      (IHe2 : UP'_P P e2),
      UP'_P P (@assign e1 e2 (proj1_sig IHe1) (proj1_sig IHe2)))
      (e : RefE (sig (UP'_P P)))
        :
        sig (UP'_P P) :=
    match e with
      | Ref e => exist (UP'_P P) _ (H _ (proj2_sig e))
      | DeRef e => exist (UP'_P P) _ (H0 _ (proj2_sig e))
      | Assign e1 e2 => exist (UP'_P P) _
        (H1 (proj1_sig e1) (proj1_sig e2) (proj2_sig e1) (proj2_sig e2))
    end.

  Lemma ind_PAlg_RefE (Name : Set)
    (P : forall e : Fix E, Universal_Property'_fold e -> Prop)
    (H : forall e
      (IHe : UP'_P P e),
      UP'_P P (@ref e (proj1_sig IHe)))
    (H0 : forall e
      (IHe : UP'_P P e),
      UP'_P P (@deref e (proj1_sig IHe)))
    (H1 : forall e1 e2
      (IHe1 : UP'_P P e1)
      (IHe2 : UP'_P P e2),
      UP'_P P (@assign e1 e2 (proj1_sig IHe1) (proj1_sig IHe2)))
    :
    PAlgebra Name RefE E (UP'_P P).
    econstructor 1 with (p_algebra := ind_alg_RefE P H H0 H1).
    destruct e; simpl.
    unfold ref, ref'; simpl; intros; rewrite wf_functor; reflexivity.
    unfold deref, deref'; simpl; intros; rewrite wf_functor; reflexivity.
    unfold assign, assign'; simpl; intros; rewrite wf_functor; reflexivity.
  Qed.

  Definition ind2_alg_RefE
    {F F' : Set -> Set}
    {Fun_F : Functor F}
    {Fun_F' : Functor F'}
    {Sub_Arith_F : RefE :<: F}
    {Sub_Arith_F' : RefE :<: F'}
    (P : forall e : (Fix F) * (Fix F'),
      Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e) -> Prop)
    (H : forall e
      (IHe : UP'_P2 P e),
      UP'_P2 P (inject (Ref _ (exist _ _ (proj1 (proj1_sig IHe)))),
      inject (Ref _ (exist _ _ (proj2 (proj1_sig IHe))))))
    (H0 : forall e
      (IHe : UP'_P2 P e),
      UP'_P2 P (inject (DeRef _ (exist _ _ (proj1 (proj1_sig IHe)))),
        inject (DeRef _ (exist _ _ (proj2 (proj1_sig IHe))))))
    (H1 : forall e1 e2
      (IHe1 : UP'_P2 P e1)
      (IHe2 : UP'_P2 P e2),
      UP'_P2 P (inject (Assign _ 
        (exist _ _ (proj1 (proj1_sig IHe1))) (exist _ _ (proj1 (proj1_sig IHe2)))),
      inject (Assign _ 
        (exist _ _ (proj2 (proj1_sig IHe1))) (exist _ _ (proj2 (proj1_sig IHe2))))))
      (e : RefE (sig (UP'_P2 P)))
        :
        sig (UP'_P2 P) :=
    match e with
      | Ref e => exist (UP'_P2 P) _ (H _ (proj2_sig e))
      | DeRef e => exist (UP'_P2 P) _ (H0 _ (proj2_sig e))
      | Assign e1 e2 => exist (UP'_P2 P) _
        (H1 (proj1_sig e1) (proj1_sig e2) (proj2_sig e1) (proj2_sig e2))
    end.

  Lemma ind_P2Alg_RefE (Name : Set)
    (F F' : Set -> Set)
    {Fun_F : Functor F}
    {Fun_F' : Functor F'}
    {Sub_RefF_F : RefE :<: F}
    {Sub_RefF_F' : RefE :<: F'}
    {WF_Fun_F : WF_Functor _ _ _ _ Fun_F}
    {WF_Fun_F' : WF_Functor _ _ _ _ Fun_F'}
    (P : forall e : (Fix F) * (Fix F'),
      Universal_Property'_fold (fst e) /\ Universal_Property'_fold (snd e) -> Prop)
    (H : forall e
      (IHe : UP'_P2 P e),
      UP'_P2 P (inject (Ref _ (exist _ _ (proj1 (proj1_sig IHe)))),
      inject (Ref _ (exist _ _ (proj2 (proj1_sig IHe))))))
    (H0 : forall e
      (IHe : UP'_P2 P e),
      UP'_P2 P (inject (DeRef _ (exist _ _ (proj1 (proj1_sig IHe)))),
        inject (DeRef _ (exist _ _ (proj2 (proj1_sig IHe))))))
    (H1 : forall e1 e2
      (IHe1 : UP'_P2 P e1)
      (IHe2 : UP'_P2 P e2),
      UP'_P2 P (inject (Assign _ 
        (exist _ _ (proj1 (proj1_sig IHe1))) (exist _ _ (proj1 (proj1_sig IHe2)))),
      inject (Assign _ 
        (exist _ _ (proj2 (proj1_sig IHe1))) (exist _ _ (proj2 (proj1_sig IHe2))))))
        :
        P2Algebra Name RefE F F' (UP'_P2 P).
    econstructor 1 with (p2_algebra := ind2_alg_RefE P H H0 H1);
      destruct e; simpl; unfold inject; simpl; 
        intros; rewrite wf_functor; reflexivity.
  Qed.

  (* Alternative Induction Principles for RefE. *)
  Definition indD_alg_RefE
    (P : forall e : Fix E, Universal_Property'_fold e -> Set)
    (H : forall e
      (IHe : UP'_SP P e),
      UP'_SP P (@ref e (projT1 IHe)))
    (H0 : forall e
      (IHe : UP'_SP P e),
      UP'_SP P (@deref e (projT1 IHe)))
    (H1 : forall e1 e2
      (IHe1 : UP'_SP P e1)
      (IHe2 : UP'_SP P e2),
      UP'_SP P (@assign e1 e2 (projT1 IHe1) (projT1 IHe2)))
      (e : RefE (sigT (UP'_SP P)))
        :
        sigT (UP'_SP P) :=
    match e with
      | Ref e => existT (UP'_SP P) _ (H _ (projT2 e))
      | DeRef e => existT (UP'_SP P) _ (H0 _ (projT2 e))
      | Assign e1 e2 => existT (UP'_SP P) _ (H1 _ _ (projT2 e1) (projT2 e2))
    end.

  Lemma ind_DAlg_RefE (Name : Set)
    (P : forall e : Fix E, Universal_Property'_fold e -> Set)
    (H : forall e
      (IHe : UP'_SP P e),
      UP'_SP P (@ref e (projT1 IHe)))
    (H0 : forall e
      (IHe : UP'_SP P e),
      UP'_SP P (@deref e (projT1 IHe)))
    (H1 : forall e1 e2
      (IHe1 : UP'_SP P e1)
      (IHe2 : UP'_SP P e2),
      UP'_SP P (@assign e1 e2 (projT1 IHe1) (projT1 IHe2)))
    :
    DAlgebra Name RefE E (UP'_SP P).
    econstructor 1 with (d_algebra := indD_alg_RefE P H H0 H1).
    destruct e; simpl.
    unfold ref, ref'; simpl; intros; rewrite wf_functor; reflexivity.
    unfold deref, deref'; simpl; intros; rewrite wf_functor; reflexivity.
    unfold assign, assign'; simpl; intros; rewrite wf_functor; reflexivity.
  Qed.

  (* ============================================== *)
  (* VALUES                                         *)
  (* ============================================== *)

   Variable V : Set -> Set.
   Context {Fun_V : Functor V}.
   Definition Value := Value V.

   Inductive LocValue (A : Set) : Set :=
   | Loc : nat -> LocValue A.

   Definition Loc_fmap : forall (A B : Set) (f : A -> B),
     LocValue A -> LocValue B :=
     fun A B _ e => match e with
                      | Loc n => Loc _ n
                    end.

   Global Instance Loc_Functor : Functor LocValue :=
     {| fmap := Loc_fmap |}.
     destruct a; reflexivity.
     destruct a; reflexivity.
   Defined.

   Context {Sub_LocValue_V : LocValue :<: V}.
   Context {Sub_BotValue_V : BotValue :<: V}.
   Context {Dis_Loc_Bot : Distinct_Sub_Functor _ Sub_LocValue_V Sub_BotValue_V}.

   (* Constructor + Universal Property. *)
   Context {WF_SubBotValue_V : WF_Functor BotValue V Sub_BotValue_V Bot_Functor Fun_V}.
   Context {WF_Sub_LocValue_F : WF_Functor _ _ Sub_LocValue_V _ _}.

   Definition loc' (n : nat) : Value := inject' (Loc _ n).
   Definition loc (n : nat) : Fix V := proj1_sig (loc' n).

   Global Instance UP'_loc {n : nat} :
     Universal_Property'_fold (loc n) := proj2_sig (loc' n).

   (* Constructor Testing for Locations. *)

   Definition isLoc : Fix V -> option nat :=
     fun exp =>
       match project exp with
         | Some (Loc n) => Some n
         | None        => None
       end.

    Lemma project_loc_loc : forall n,
      project (F := V) (G := LocValue) (loc n) = Some (Loc _ n).
    Proof.
      intros; unfold project, loc, inject; simpl; rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl; unfold Loc_fmap.
      rewrite prj_inj; reflexivity.
    Qed.

    (* Unit Values *)

   Inductive UnitValue (A : Set) : Set :=
   | Unit : UnitValue A.

   Definition Unit_fmap : forall (A B : Set) (f : A -> B),
     UnitValue A -> UnitValue B :=
     fun A B _ e => match e with
                      | Unit => Unit B
                    end.

   Global Instance Unit_Functor : Functor UnitValue :=
     {| fmap := Unit_fmap |}.
     destruct a; reflexivity.
     destruct a; reflexivity.
   Defined.

   Context {Sub_UnitValue_V : UnitValue :<: V}.
   Context {Dis_Unit_Bot : Distinct_Sub_Functor _ Sub_UnitValue_V Sub_BotValue_V}.

   (* Constructor + Universal Property. *)
   Context {WF_Sub_UnitValue_F : WF_Functor _ _ Sub_UnitValue_V _ _}.

   Definition unit' : Value := inject' (Unit _).
   Definition unit : Fix V := proj1_sig unit'.

   Global Instance UP'_unit :
     Universal_Property'_fold unit := proj2_sig unit'.

   (* Constructor Testing for Unitations. *)

   Definition isUnit : Fix V -> bool :=
     fun exp =>
       match project exp with
         | Some Unit => true
         | None      => false
       end.

    Lemma project_unit_unit : project (F := V) (G := UnitValue) unit = Some (Unit _).
    Proof.
      intros; unfold project, unit, inject; simpl; rewrite out_in_fmap.
      repeat rewrite wf_functor; simpl; unfold Unit_fmap.
      rewrite prj_inj; reflexivity.
    Qed.

   (* Inversion principles for natural number SubValues. *)

   Variable SV : (SubValue_i V -> Prop) -> SubValue_i V -> Prop.
   Context {iFun_SV : iFunctor SV}.
   Context {Sub_SV_refl_SV : Sub_iFunctor (SubValue_refl V) SV}.

   Definition SV_invertLoc_P (i : SubValue_i V) :=
     forall n, proj1_sig (sv_a _ i) = loc n ->
       proj1_sig (sv_b _ i) = loc n.

    Inductive SV_invertLoc_Name := ece_invertloc_name.
    Context {SV_invertLoc_SV :
      iPAlgebra SV_invertLoc_Name SV_invertLoc_P SV}.

    Global Instance SV_invertLoc_refl :
      iPAlgebra SV_invertLoc_Name SV_invertLoc_P (SubValue_refl V).
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_invertLoc_P.
      inversion H; subst; simpl; congruence.
    Defined.

    Lemma SV_invertLoc_default : forall V'
      (Fun_V' : Functor V')
      (SV' : (SubValue_i V -> Prop) -> SubValue_i V -> Prop)
      (sub_V'_V : V' :<: V)
      (WF_V' : WF_Functor V' V sub_V'_V Fun_V' Fun_V),
      (forall (i : SubValue_i V) (H : SV' SV_invertLoc_P i),
        exists v', proj1_sig (sv_a _ i) = inject v') ->
      Distinct_Sub_Functor _ Sub_LocValue_V sub_V'_V ->
      iPAlgebra SV_invertLoc_Name SV_invertLoc_P SV'.
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_invertLoc_P.
      destruct (H _ H1) as [v' eq_v'].
      intros; rewrite eq_v' in H2.
      elimtype False.
      unfold loc, inject, loc', inject' in H2; simpl in H2.
      apply sym_eq in H2.
      apply (inject_discriminate H0 _ _ H2).
    Defined.

    Global Instance SV_invertLoc_Bot :
      iPAlgebra SV_invertLoc_Name SV_invertLoc_P (SubValue_Bot V).
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_invertLoc_P.
      inversion H; subst; simpl; intros.
      elimtype False.
      rewrite H0 in H1.
      unfold loc, inject, loc', inject' in H1; simpl in H1.
      repeat rewrite out_in_inverse, wf_functor in H1; simpl in H1.
      eapply (inject_discriminate Dis_Loc_Bot); unfold inject; simpl; eauto.
    Defined.

    Context {iFun_F : iFunctor SV}.
    Definition SV_invertLoc := ifold_ SV _ (ip_algebra (iPAlgebra := SV_invertLoc_SV)).

    Definition SV_invertLoc'_P (i : SubValue_i V) :=
      forall n, proj1_sig (sv_b _ i) = loc n ->
        proj1_sig (sv_a _ i) = loc n \/ proj1_sig (sv_a _ i) = bot _.

    Inductive SV_invertLoc'_Name := ece_invertloc'_name.
    Context {SV_invertLoc'_SV :
      iPAlgebra SV_invertLoc'_Name SV_invertLoc'_P SV}.

    Global Instance SV_invertLoc'_refl :
      iPAlgebra SV_invertLoc'_Name SV_invertLoc'_P (SubValue_refl V).
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_invertLoc'_P.
      inversion H; subst; simpl; eauto.
      intros.
      left; congruence.
    Defined.

    Global Instance SV_invertLoc'_Bot :
      iPAlgebra SV_invertLoc'_Name SV_invertLoc'_P (SubValue_Bot V).
      econstructor; intros.
      unfold iAlgebra; intros; unfold SV_invertLoc'_P.
      inversion H; subst; simpl; eauto.
    Defined.

    Definition SV_invertLoc' := ifold_ SV _ (ip_algebra (iPAlgebra := SV_invertLoc'_SV)).

   Context {Sub_StuckValue_V : StuckValue :<: V}.
   Definition stuck' : nat -> Value := stuck' _.
   Definition stuck : nat -> Fix V := stuck _.

  (* ============================================== *)
  (* TYPING                                         *)
  (* ============================================== *)

  (* Typing Reference Expressions. *)
  
  Variable (MT : Set -> Set).
  Context {Fun_MT : Functor MT}.
  Context {Mon_MT : Monad MT}.
  Context {Fail_MT : FailMonad MT}.

  Definition RefE_typeof (R : Set) (rec : R -> typeofR D MT)
     (e : RefE R) : typeofR D MT :=
     match e with
       | Ref e => do t <- rec e;
                  return_ (tref' t)
       | DeRef e => do te <- rec e;
                   match isTRef (proj1_sig te) with
                     | Some t => return_ t
                     | None => fail
                   end
       | Assign e1 e2 => do t1 <- rec e1;
                           match isTRef (proj1_sig t1) with
                             | Some t =>
                               do t2 <- rec e2;
                                 if (eq_DType D (proj1_sig t) t2) then
                                   return_ tunit'
                                   else
                                     fail
                             | None => fail
                           end
     end.

  Global Instance MAlgebra_typeof_RefE T:
    FAlgebra TypeofName T (typeofR D MT) RefE :=
    {| f_algebra := RefE_typeof T|}.

  (* ============================================== *)
  (* EVALUATION                                     *)
  (* ============================================== *)

  Variable (ME : Set -> Set). (* Evaluation Monad. *)
  Context {Fun_M : Functor ME}.
  Context {Mon_M : Monad ME}.
  Context {FailM : FailMonad ME}.
  Context {ExceptionM : Exception ME nat}.
  Context {State_M : StateM ME (list Value)}.

  (* Monadic Evaluation Algebra for RefEemetic Expressions. *)

  Definition RefE_evalM (R : Set) (rec : R -> evalMR V ME)
     (e : RefE R) : evalMR V ME :=
       match e with
         | Ref e =>
           do v <- (rec e);
             do env <- get;
               put (insert _ v env) >> return_ (loc' (length env))
         | DeRef e =>
           do v <- (rec e);
             match isLoc (proj1_sig v) with
               | Some n => do env <- get;
                 return_ match (@lookup Value env n) with
                           | Some mv => mv
                           | _ => stuck' 457
                         end
               | None => return_ (stuck' 456)
             end
         | Assign e1 e2 =>
           do v <- (rec e1);
             match isLoc (proj1_sig v) with
               | Some n =>
                 do v2 <- rec e2;
                   do env <- get;
                     put (replace_el n v2 env) >> return_ unit'
               | None => return_ (stuck' 3)
             end
       end.

  Global Instance MAlgebra_evalM_RefE T :
    FAlgebra EvalName T (evalMR V ME) RefE :=
    {| f_algebra := RefE_evalM T|}.

  (* ============================================== *)
  (* PRETTY PRINTING                                *)
  (* ============================================== *)

  (* Pretty Printing Functions*)

  Global Instance MAlgebra_ExpPrint_RefE T :
    FAlgebra ExpPrintName T (ExpPrintR) RefE :=
    {| f_algebra := fun rec e =>
      match e with
        | Ref e => fun i => append "ref (" ((rec e i) ++ ")")
        | DeRef e => fun i => append "! (" ((rec e i) ++ ")")
        | Assign e1 e2 => fun i => append (rec e1 i) (" := " ++ (rec e2 i))
    end |}.

  Global Instance MAlgebra_ValuePrint_Loc T :
    FAlgebra ValuePrintName T ValuePrintR LocValue :=
    {| f_algebra := fun rec e =>
      match e with
        | Loc n => append "loc_" (String (ascii_of_nat (n + 48)) EmptyString)
      end |}.

  (* ============================================== *)
  (* CONTINUITY Of MONADIC EVALUATION               *)
  (* ============================================== *)

  Context {evalM_E : FAlgebra EvalName Exp (evalMR V ME) E}.
  Context {WF_eval_F : @WF_FAlgebra EvalName _ _ RefE E
    Sub_RefE_E (MAlgebra_evalM_RefE _) evalM_E}.

  (* Continuity of Monadic Evaluation. *)

  Variable SVM : (SubValueM_i V ME -> Prop) -> SubValueM_i V ME -> Prop.
  Context {iFun_SVM : iFunctor SVM}.
  Context {Sub_SVM_Base_SVM : Sub_iFunctor (SubValueM_base V ME) SVM}.

  (* Definition WF_Env := P2_Env (SubValueC V SV). *)

  Inductive SubValueM_State (A : SubValueM_i V ME -> Prop) : SubValueM_i V ME -> Prop :=
  | SVM_Get : forall k k' : list Value -> ME Value,
    (forall env, A (mk_SubValueM_i V ME (k env) (k' env))) ->
    SubValueM_State A (mk_SubValueM_i V ME (do x <- (MonadLib.get (M := ME)); k x)
      (do y <- (MonadLib.get (M := ME)); k' y))
  | SVM_Put : forall env j j',
    A (mk_SubValueM_i V ME j j') ->
    SubValueM_State A (mk_SubValueM_i V ME (put env >> j) (put env >> j')).

  Definition ind_alg_SVM_State (P : SubValueM_i V ME -> Prop)
    (H : forall k k' : list Value -> ME Value,
      (forall env, P (mk_SubValueM_i V ME (k env) (k' env))) ->
      P (mk_SubValueM_i V ME (do x <- (MonadLib.get (M := ME)); k x)
        (do y <- (MonadLib.get (M := ME)); k' y)))
    (H0 : forall env j j',
      P (mk_SubValueM_i V ME j j') ->
      P (mk_SubValueM_i V ME (put env >> j) (put env >> j')))
    i (e : SubValueM_State P i) : P i :=
    match e in SubValueM_State _ i return P i with
      | SVM_Get k k' Sub_k_k' => H k k' Sub_k_k'
      | SVM_Put env j j' Sub_j_j' => H0 env j j' Sub_j_j'
    end.

  Definition SubValueM_State_ifmap (A B : SubValueM_i V ME-> Prop) i (f : forall i, A i -> B i)
    (SVM_a : SubValueM_State A i) : SubValueM_State B i :=
    match SVM_a in (SubValueM_State _ s) return (SubValueM_State B s) with
      | SVM_Get k k' Sub_k_k' =>
          SVM_Get _ k k' (fun env => f _ (Sub_k_k' env))
      | SVM_Put env j j' Sub_j_j' =>
        SVM_Put _ env j j' (f _ Sub_j_j')
    end.

  Global Instance iFun_SubValueM_State : iFunctor SubValueM_State.
  Proof.
    constructor 1 with (ifmap := SubValueM_State_ifmap).
    destruct a; simpl; intros; reflexivity.
    destruct a; simpl; intros.
    apply f_equal; repeat (apply functional_extensionality_dep; intros);
      unfold id; auto.
    unfold id; auto.
  Defined.

  Lemma ind_iPAlg_SVM_State (Name : Set)
    (P : SubValueM_i V ME -> Prop)
    (H : forall k k' : list Value -> ME Value,
      (forall env, P (mk_SubValueM_i V ME (k env) (k' env))) ->
      P (mk_SubValueM_i V ME (do x <- (MonadLib.get (M := ME)); k x)
        (do y <- (MonadLib.get (M := ME)); k' y)))
    (H0 : forall env j j',
      P (mk_SubValueM_i V ME j j') ->
      P (mk_SubValueM_i V ME (put env >> j) (put env >> j')))
    :
    iPAlgebra Name P SubValueM_State.
  Proof.
    econstructor; unfold iAlgebra; intros; apply ind_alg_SVM_State; auto.
  Qed.

  Variable Sub_SVM_State_SVM : Sub_iFunctor SubValueM_State SVM. 

    Global Instance SVM_Bind_State : 
      iPAlgebra SVM_Bind_Name (SVM_Bind_P V ME SVM) SubValueM_State.
    Proof.
      econstructor; unfold iAlgebra; intros; apply ind_alg_SVM_State; 
        eauto; unfold SVM_Bind_P; simpl; intros.
      (* SVM_Get *)
      repeat rewrite <- associativity; apply (inject_i (subGF := Sub_SVM_State_SVM));
        constructor 1; intros; eapply H0; eauto.
      (* SVM_Return *)
      unfold wbind; repeat rewrite <- associativity; apply (inject_i (subGF := Sub_SVM_State_SVM));
        constructor 2; intros; eapply H0; eauto.
    Qed.

  (* Ref case. *)
  Ltac WF_FAlg_rewrite := repeat rewrite wf_functor; simpl;
    repeat rewrite out_in_fmap; simpl;
      repeat rewrite wf_functor; simpl;
        repeat rewrite wf_algebra; simpl.

  Context {SVM_Bind_SVM: iPAlgebra SVM_Bind_Name (SVM_Bind_P V ME SVM) SVM}.

    Lemma bevalM_continuous_Exp_H : forall
      (e : Fix E)
      (IHe : UP'_SP (bevalM_continuous_Exp_P V E ME SVM) e),
      UP'_SP (bevalM_continuous_Exp_P V E ME SVM) (@ref e (projT1 IHe)).
    Proof.
      unfold bevalM_continuous_Exp_P; intros.
      destruct IHe as [e_UP' IHe].
      econstructor; intros; eauto with typeclass_instances.
      unfold Names.bevalM, mfold; simpl; rewrite H1; unfold ref; simpl.
      WF_FAlg_rewrite.
      repeat erewrite bF_UP_in_out.
      generalize (H (exist Universal_Property'_fold e e_UP') _ _ H0 (refl_equal _))
        as SVM_e_e'; intros.
      apply SVM_Bind; eauto with typeclass_instances.
      intros; apply (inject_i (subGF := Sub_SVM_State_SVM)); 
        econstructor; intros; apply (inject_i (subGF := Sub_SVM_State_SVM)).
        econstructor; intros; apply (inject_i (subGF := Sub_SVM_Base_SVM)); constructor; auto.
    Qed.

    (* DeRef case. *)

    Lemma bevalM_continuous_Exp_H0 : forall
      (e : Fix E)
      (IHe : UP'_SP (bevalM_continuous_Exp_P V E ME SVM) e),
      UP'_SP (bevalM_continuous_Exp_P V E ME SVM) (@deref e (projT1 IHe)).
    Proof.
      unfold bevalM_continuous_Exp_P; intros.
      destruct IHe as [e_UP' IHe].
      econstructor; intros; eauto with typeclass_instances.
      unfold Names.bevalM, mfold; simpl; rewrite H1; unfold deref; simpl.
      WF_FAlg_rewrite.
      repeat erewrite bF_UP_in_out.
      generalize (H (exist Universal_Property'_fold e e_UP') _ _ H0 (refl_equal _))
        as SVM_e_e'; intros.
      apply SVM_Bind; eauto with typeclass_instances.
      intros; caseEq (isLoc (proj1_sig v)). 
      apply (inject_i (subGF := Sub_SVM_State_SVM)); 
        econstructor; intros; apply (inject_i (subGF := Sub_SVM_Base_SVM));
          constructor; auto.
      apply (inject_i (subGF := Sub_SVM_Base_SVM)); constructor; auto.
    Qed.

    (* Assign Case *)

    Lemma bevalM_continuous_Exp_H1 : forall
      (m n : Fix E)
      (IHm : UP'_SP (bevalM_continuous_Exp_P V E ME SVM) m)
      (IHn : UP'_SP (bevalM_continuous_Exp_P V E ME SVM) n),
      UP'_SP (bevalM_continuous_Exp_P V E ME SVM) (@assign m n (projT1 IHm) (projT1 IHn)).
      unfold bevalM_continuous_Exp_P; intros.
      destruct IHm as [m_UP' IHm].
      destruct IHn as [n_UP' IHn].
      econstructor; intros; eauto with typeclass_instances.
      unfold Names.bevalM, mfold; simpl; rewrite H1; unfold assign; simpl.
      WF_FAlg_rewrite.
      repeat erewrite bF_UP_in_out.
      generalize (H (exist Universal_Property'_fold m m_UP') _ _ H0 (refl_equal _))
        as SVM_m_m'; intros.
      generalize (H (exist Universal_Property'_fold n n_UP') _ _ H0 (refl_equal _))
        as SVM_n_n'; intros.
      apply SVM_Bind; eauto with typeclass_instances.
      intros; caseEq (isLoc (proj1_sig v)). 
      apply SVM_Bind; eauto with typeclass_instances.
      intros; apply (inject_i (subGF := Sub_SVM_State_SVM)); 
        econstructor; intros; apply (inject_i (subGF := Sub_SVM_State_SVM));
        econstructor; intros; apply (inject_i (subGF := Sub_SVM_Base_SVM));
          constructor; auto.
      apply (inject_i (subGF := Sub_SVM_Base_SVM)); constructor; auto.
    Qed.

    Global Instance RefE_bevalM_continuous_Exp  :
      DAlgebra EC_ExpName RefE E (UP'_SP (bevalM_continuous_Exp_P V E ME SVM)).
    Proof.
      apply ind_DAlg_RefE.
      apply bevalM_continuous_Exp_H.
      apply bevalM_continuous_Exp_H0.
      apply bevalM_continuous_Exp_H1.
    Qed.

  Section WFValue_Sec.

  (* ============================================== *)
  (* WELL-FORMED REFERENCE VALUES                   *)
  (* ============================================== *)

    Variable TypContext : Set.
    Context {TypContextCE : ConsExtensionC TypContext}.
    
    (** Typing rules for unit values **)

    Inductive WFValue_Unit (WFV : WFValue_i D V TypContext -> Prop) : WFValue_i D V TypContext -> Prop :=
    | WFVS_Unit : forall Sigma v T,
      proj1_sig v = unit ->
      proj1_sig T = tunit ->
      WFValue_Unit WFV (mk_WFValue_i D V _ Sigma v T).

    Definition ind_alg_WFV_Unit (P : WFValue_i D V TypContext -> Prop)
      (H : forall Sigma v T v_eq T'_eq, P (mk_WFValue_i D V _ Sigma v T))
      i (e : WFValue_Unit P i) : P i :=
      match e in WFValue_Unit _ i return P i with
        | WFVS_Unit Sigma v T v_eq T_eq =>
          H Sigma v T v_eq T_eq
      end.

    Definition WFV_Unit_ifmap (A B : WFValue_i D V TypContext -> Prop) i (f : forall i, A i -> B i)
      (WFVS_a : WFValue_Unit A i) : WFValue_Unit B i :=
      match WFVS_a in (WFValue_Unit _ s) return (WFValue_Unit B s)
        with
        | WFVS_Unit Sigma v T v_eq T_eq =>
          WFVS_Unit _ Sigma v T v_eq T_eq
      end.

    Global Instance iFun_WFVS_Unit : iFunctor WFValue_Unit.
      constructor 1 with (ifmap := WFV_Unit_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; reflexivity.
    Defined.

    Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
    Variable funWFV : iFunctor WFV.
    Variable Sub_WFV_Unit_WFV : Sub_iFunctor WFValue_Unit WFV.

    Global Instance WFV_proj1_a_Unit :
      iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V _ WFV) WFValue_Unit.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WFV_proj1_a_P.
      inversion H; subst; simpl; intros.
      apply (inject_i (subGF := Sub_WFV_Unit_WFV)); econstructor; simpl; eauto.
      rewrite H3; eauto.
    Defined.

    Global Instance WFV_proj1_b_Unit :
      iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV) WFValue_Unit.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WFV_proj1_b_P.
      inversion H; subst; simpl; intros.
      apply (inject_i (subGF := Sub_WFV_Unit_WFV)); econstructor; simpl; eauto.
      rewrite H3; eauto.
    Defined.

    Global Instance WFV_Weaken_Unit :
      iPAlgebra WFValue_Weaken_Name (WFValue_Weaken_P D V _ WFV) WFValue_Unit.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WFValue_Weaken_P.
      inversion H; subst; simpl; intros.
      apply inject_i; econstructor; eassumption.
    Defined.

    (* Inversion principles for well-formed unit values. *)

    Definition WF_invertUnit_P (i : WFValue_i D V _) :=
      forall T' T'_UP,
        proj1_sig (wfv_T _ _ _ i) = @tref T' T'_UP ->
        WFValue_Unit (iFix WFV) i.

    Inductive WF_invertUnit_Name := wfv_invertunit_name.
    Context {WF_invertUnit_WFV :
      iPAlgebra WF_invertUnit_Name WF_invertUnit_P WFV}.

    Global Instance WF_invertUnit_Unit :
      iPAlgebra WF_invertUnit_Name WF_invertUnit_P WFValue_Unit.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertUnit_P.
      inversion H; subst; simpl; intros.
      econstructor; eassumption.
    Defined.

    Definition WF_invertUnit := ifold_ WFV _ (ip_algebra (iPAlgebra := WF_invertUnit_WFV)).

    (** Typing rules for location values **)

    Class SigTypContextC (TypContext : Set) : Set :=
      {SigLookup : TypContext -> nat -> option DType;
       SigInsert : DType -> TypContext -> TypContext}.

    Class SigConsExtensionC (TypContext : Set) 
      (ConsExt : ConsExtensionC TypContext)
      (SigTyp : SigTypContextC TypContext) : Type :=
      { 
        ConsExtension_SigLookup : forall Sigma Sigma',
          ConsExtension Sigma' Sigma -> 
          forall n T, SigLookup Sigma n = Some T -> 
            SigLookup Sigma' n = Some T;
        ConsExtension_SigInsert : forall Sigma T,
          ConsExtension (SigInsert T Sigma) Sigma}.

    Context {TypContext_S : SigTypContextC TypContext}.
    Context {TypContext_SCE : SigConsExtensionC TypContext _ _}.

    Inductive WFValue_Loc (WFV : WFValue_i D V TypContext -> Prop) : WFValue_i D V TypContext -> Prop :=
    | WFVS_Loc : forall Sigma v n (T T' : DType),
      SigLookup Sigma n = Some T' ->
      proj1_sig v = loc n ->
      proj1_sig T = @tref _ (proj2_sig T') ->
      WFValue_Loc WFV (mk_WFValue_i D V _ Sigma v T).

    Definition ind_alg_WFV_Loc (P : WFValue_i D V _ -> Prop)
      (H : forall Sigma v n T T' lookup_n v_eq T'_eq, P (mk_WFValue_i D V _ Sigma v T))
      i (e : WFValue_Loc P i) : P i :=
      match e in WFValue_Loc _ i return P i with
        | WFVS_Loc Sigma v n T T' lookup_n v_eq T_eq =>
          H Sigma v n T T' lookup_n v_eq T_eq
      end.

    Definition WFV_Loc_ifmap (A B : WFValue_i D V _ -> Prop) i (f : forall i, A i -> B i)
      (WFVS_a : WFValue_Loc A i) : WFValue_Loc B i :=
      match WFVS_a in (WFValue_Loc _ s) return (WFValue_Loc B s)
        with
        | WFVS_Loc Sigma v n T T' lookup_n v_eq T_eq =>
          WFVS_Loc _ Sigma v n T T' lookup_n v_eq T_eq
      end.

    Global Instance iFun_WFVS_Loc : iFunctor WFValue_Loc.
      constructor 1 with (ifmap := WFV_Loc_ifmap).
      destruct a; simpl; intros; reflexivity.
      destruct a; simpl; intros; reflexivity.
    Defined.

    Variable Sub_WFV_Loc_WFV : Sub_iFunctor WFValue_Loc WFV.

    Global Instance WFV_Weaken_Loc :
      iPAlgebra WFValue_Weaken_Name (WFValue_Weaken_P D V _ WFV) WFValue_Loc.
    Proof.
      constructor; intros.
      unfold iAlgebra; intros; apply ind_alg_WFV_Loc; try eassumption;
        unfold WFValue_Weaken_P; simpl; intros.
      apply inject_i; econstructor; eauto.
      eapply ConsExtension_SigLookup; eauto.
    Qed.

    Global Instance WFV_proj1_a_Loc :
      iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V _ WFV) WFValue_Loc.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WFV_proj1_a_P.
      inversion H; subst; simpl; intros.
      apply (inject_i (subGF := Sub_WFV_Loc_WFV)); econstructor; simpl; eauto.
      rewrite H4; eauto.
    Defined.

    Global Instance WFV_proj1_b_Loc :
      iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV) WFValue_Loc.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WFV_proj1_b_P.
      inversion H; subst; simpl; intros.
      apply (inject_i (subGF := Sub_WFV_Loc_WFV)); econstructor; simpl; eauto.
      rewrite H4; eauto.
    Defined.

    (* Inversion principles for well-formed location values. *)

    Definition WF_invertLoc_P (i : WFValue_i D V _) :=
      forall T' T'_UP,
        proj1_sig (wfv_T _ _ _ i) = @tref T' T'_UP ->
        WFValue_Loc (iFix WFV) i.

    Inductive WF_invertLoc_Name := wfv_invertloc_name.
    Context {WF_invertLoc_WFV :
      iPAlgebra WF_invertLoc_Name WF_invertLoc_P WFV}.

    Global Instance WF_invertLoc_Loc :
      iPAlgebra WF_invertLoc_Name WF_invertLoc_P WFValue_Loc.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertLoc_P.
      inversion H; subst; simpl; intros.
      econstructor; eassumption.
    Defined.

    Context {Dis_TRef_TUnit : Distinct_Sub_Functor _ Sub_RefType_D Sub_UnitType_D}.

    Global Instance WF_invertLoc_Unit :
      iPAlgebra WF_invertLoc_Name WF_invertLoc_P WFValue_Unit.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertLoc_P.
      inversion H; subst; simpl; intros.
      rewrite H2 in H1.
    unfold tunit, tref, tunit', tref' in H1.
    elimtype False.
    generalize (inject_discriminate Dis_TRef_TUnit
      (TRef (sig Universal_Property'_fold) (exist _ T' T'_UP)) (TUnit (sig Universal_Property'_fold))).
    revert H1; unfold inject, inject'; eauto.
    Defined.

    Definition WF_invertLoc := ifold_ WFV _ (ip_algebra (iPAlgebra := WF_invertLoc_WFV)).

    Definition WF_invertLoc'_P (i : WFValue_i D V _) :=
      forall n,
        proj1_sig (wfv_v _ _ _ i) = loc n ->
        WFValue_Loc (iFix WFV) i.

    Inductive WF_invertLoc'_Name := wfv_invertloc'_name.
    Context {WF_invertLoc'_WFV :
      iPAlgebra WF_invertLoc'_Name WF_invertLoc'_P WFV}.

    Global Instance WF_invertLoc'_Loc :
      iPAlgebra WF_invertLoc'_Name WF_invertLoc'_P WFValue_Loc.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertLoc'_P.
      inversion H; subst; simpl; intros.
      econstructor; eassumption.
    Defined.

    Context {Dis_LocValue_UnitValue : Distinct_Sub_Functor _ Sub_LocValue_V Sub_UnitValue_V}.

    Global Instance WF_invertLoc'_Unit :
      iPAlgebra WF_invertLoc'_Name WF_invertLoc'_P WFValue_Unit.
    Proof.
      econstructor; intros.
      unfold iAlgebra; intros; unfold WF_invertLoc'_P.
      inversion H; subst; simpl; intros.
      rewrite H2 in H0.
      unfold tunit, tref, tunit', tref' in H0.
      elimtype False.
      generalize (inject_discriminate Dis_LocValue_UnitValue
        (Loc _ n) (Unit (sig Universal_Property'_fold))).
      revert H0; unfold inject, inject'; eauto.
    Defined.

    Definition WF_invertLoc' := ifold_ WFV _ (ip_algebra (iPAlgebra := WF_invertLoc'_WFV)).

  (* ============================================== *)
  (* WELL-FORMED MONADIC STATE VALUES               *)
  (* ============================================== *)

    (** SuperFunctor for Well-Formed Monadic Value Relation with
       Typing Environment. **)

    Class WF_EnvC (TypContext : Set) : Type := 
      {WF_Env : list Value -> TypContext -> Prop}.

    Context {TypContext_WFE : WF_EnvC TypContext}.

    Inductive WFValueM_State (A : WFValueM_i D V MT ME TypContext -> Prop) : WFValueM_i D V MT ME TypContext -> Prop :=
    | WFVM_Get : forall Sigma (k : list Value -> ME Value) T,
      (forall env, WF_Env env Sigma ->
        A (mk_WFValueM_i D V MT ME _ Sigma (k env) T)) ->
      WFValueM_State A (mk_WFValueM_i D V MT ME _ Sigma (do env <- get; k env) T)
    | WFVM_Put : forall Sigma Sigma' env k T,
      WF_Env env Sigma' -> ConsExtension Sigma' Sigma -> 
      A (mk_WFValueM_i _ _ _ _ _ Sigma' k T) ->
      WFValueM_State A (mk_WFValueM_i _ _ _ _ _ Sigma (put env >> k) T).

    Definition ind_alg_WFVM_State (P : WFValueM_i D V MT ME TypContext -> Prop)
      (H_Get : forall Sigma k T,
        (forall env, WF_Env env Sigma ->
          P (mk_WFValueM_i _ _ _ _ _ Sigma (k env) T)) ->
        P (mk_WFValueM_i _ _ _ _ _ Sigma (do env <- get; k env) T))
      (H_Put : forall Sigma Sigma' env k T,
        WF_Env env Sigma' -> ConsExtension Sigma' Sigma -> 
        P (mk_WFValueM_i _ _ _ _ _ Sigma' k T) ->
        P (mk_WFValueM_i _ _ _ _ _ Sigma (put env >> k) T))
      (i : WFValueM_i D V MT ME _)
      (e : WFValueM_State P i)
      :
      P i :=
      match e in WFValueM_State _ i return P i with
        | WFVM_Get Sigma k T WF_k =>
          H_Get Sigma k T WF_k
        | WFVM_Put Sigma Sigma' env k T WF_env Sub_Sig_Sig' WF_k =>
          H_Put Sigma Sigma' env k T WF_env Sub_Sig_Sig' WF_k
      end.

    Definition WFVM_State_ifmap
      (A B : WFValueM_i D V MT ME _ -> Prop)
      (i : WFValueM_i D V MT ME _)
      (f : forall i, A i -> B i)
      (WFVM_a : WFValueM_State A i)
      :
      WFValueM_State B i :=
      match WFVM_a in WFValueM_State _ i return WFValueM_State B i with
        | WFVM_Get Sigma k T WF_k =>
          WFVM_Get B Sigma k T (fun env WF_env => f _ (WF_k env WF_env))
        | WFVM_Put Sigma Sigma' env k T WF_env Sub_Sig_Sig' WF_k =>
          WFVM_Put B Sigma Sigma' env k T WF_env Sub_Sig_Sig' (f _ WF_k)
      end.

    Global Instance iFun_wfvm_State : iFunctor WFValueM_State.
      constructor 1 with (ifmap := WFVM_State_ifmap).
      (* ifmap_fusion *)
      destruct a; simpl; intros; reflexivity.
      (* ifmap_id *)
      destruct a; simpl; intros; auto;
        apply f_equal; repeat (apply functional_extensionality_dep; intros); auto.
    Defined.

    End WFValue_Sec.
    
    Variable TypContext : Set.

    Class Sig_WF_EnvC (TypContext : Set)
      (WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop)
      (WF_Env' : WF_EnvC TypContext)
      (SigTyp : SigTypContextC TypContext) : Type :=
    {  
      WF_EnvLookup : forall env Sigma T n,
        WF_Env env Sigma -> 
        SigLookup Sigma n = Some T ->
        exists v, lookup env n = Some v /\
          WFValueC _ _ _ WFV Sigma v T;
      WF_EnvInsertLookup : forall env Sigma T,
        WF_Env env Sigma -> 
        SigLookup (SigInsert T Sigma) (length env) = Some T;
      WF_EnvInsert : forall Sigma env v T,
        WFValueC _ _ _ WFV (SigInsert T Sigma) v T -> 
        WF_Env env Sigma -> 
        WF_Env (insert _ v env) (SigInsert T Sigma);
      WF_EnvReplace : forall Sigma env v T n,
        SigLookup Sigma n = Some T -> 
        WFValueC _ _ _ WFV Sigma v T -> 
        WF_Env env Sigma -> 
        WF_Env (replace_el n v env) Sigma}.

    Context {TypContextCE : ConsExtensionC TypContext}.
    Context {TypContext_S : SigTypContextC TypContext}.
    Context {TypContext_SCE : SigConsExtensionC TypContext _ _}.
    Context {TypContext_WFE : WF_EnvC TypContext}.

    Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
    Variable funWFV : iFunctor WFV.

    Context {TypContext_SWFE : Sig_WF_EnvC TypContext WFV _ _}.

    Context {Sub_WFV_Unit_WFV : Sub_iFunctor (WFValue_Unit _) WFV}.
    Context {Sub_WFV_Loc_WFV : Sub_iFunctor (WFValue_Loc _) WFV}.

    Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> WFValueM_i D V MT ME TypContext -> Prop.
    Variable funWFVM : iFunctor WFVM.

    Context {Sub_WFVM_State_WFVM : Sub_iFunctor (WFValueM_State TypContext) WFVM}.

    Context {WFV_proj1_a_WFV : iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V _ WFV) WFV}.
    Context {WFV_proj1_b_WFV : iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV) WFV}.

    Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR D MT) E}.
    Context {WF_Typeof_F : forall T, @WF_FAlgebra TypeofName _ _ RefE E
      Sub_RefE_E (MAlgebra_typeof_RefE _) (Typeof_F T)}.

    Context {evalM_E' : forall T, FAlgebra EvalName T (evalMR V ME) E}.
    Context {WF_eval_E' : forall T, @WF_FAlgebra EvalName _ _ RefE E
      Sub_RefE_E (MAlgebra_evalM_RefE _) (evalM_E' T)}.

    Context {WFV_Weaken_WFV : iPAlgebra WFValue_Weaken_Name (WFValue_Weaken_P D V _ WFV) WFV}.

    Definition WFV_Weaken := ifold_ WFV _ (ip_algebra (iPAlgebra := WFV_Weaken_WFV)).

    Definition eval_Soundness_Ref_P (i : WFValueM_i D V MT ME TypContext) :=
      WFValueM _ _ _ _ _ WFVM i /\
        WFValueMC _ _ _ _ _ WFVM (wfvm_S _ _ _ _ _ i) 
        (do v <- wfvm_v _ _ _ _ _ i;
          do env <- get;
            put (insert (Names.Value V) v env) >>
            return_ (loc' (Datatypes.length env)))
        (do t <- wfvm_T _ _ _ _ _ i;
          return_ (tref' t)).

    Inductive eval_Soundness_Ref_Name := eval_soundness_ref_name.

    Context {Sub_WFVM_Base_WFVM : Sub_iFunctor (WFValueM_base D V MT ME TypContext WFV) WFVM}.

    Global Instance wfvm_Base_eval_Soundness_Ref :
      iPAlgebra eval_Soundness_Ref_Name eval_Soundness_Ref_P (WFValueM_base D V MT ME TypContext WFV).
    Proof.
      constructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (WFV := WFV)
        (ME := ME) (Monad_ME := Mon_M) (Fail_MT := Fail_MT);
        try eassumption; unfold eval_Soundness_Ref_P; split; simpl; intros.
      (* WFVM_Return *)
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor; eauto.
      repeat rewrite <- left_unit.
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; intros.
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)).
      destruct H1 as [mt' mt'_eq].
      econstructor 2 with (Sigma' := SigInsert T Sigma).
      rename H2 into WF_env_Sigma'.
      apply WF_EnvInsert; eauto.
      eapply (WFV_Weaken _ H0); simpl; eauto.
      eapply ConsExtension_SigInsert; eauto.
      eapply ConsExtension_SigInsert; eauto.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)).
      constructor 1 with (T := tref' T).
      apply inject_i; econstructor 1 with (T' := T);
        try unfold loc; try (unfold tref; destruct T as [T T_UP']; simpl); eauto.
      eapply WF_EnvInsertLookup; eauto.
      rewrite <- mt'_eq; apply bind_orn; intros.
      destruct a as [T' T'_eq]; simpl; apply return_orn.
      unfold tref'; simpl.
      repeat rewrite wf_functor; simpl.
      rewrite T'_eq; auto.
      (* WFVM_Untyped' *)
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor; eauto.
      unfold wbind; rewrite <- associativity, bind_fail.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)).
      constructor.
    Qed.

    Global Instance WFVM_State'_eval_Soundness_Ref :
       iPAlgebra eval_Soundness_Ref_Name eval_Soundness_Ref_P (WFValueM_State TypContext).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
        try eassumption; unfold eval_Soundness_Ref_P; split; simpl; intros.
      (* WFVM_Get *)
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      intros env WF_env; apply (proj1 (H0 env WF_env)).
      rewrite <- associativity; apply (inject_i (subGF := Sub_WFVM_State_WFVM));
        econstructor; intros; eapply H0; eauto.
      (* WFVM_Put *)
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      apply (proj1 H2).
      unfold wbind; rewrite <- associativity.
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      apply (proj2 H2).
    Qed.

    Context {eval_Soundness_Ref_WFVM : iPAlgebra eval_Soundness_Ref_Name eval_Soundness_Ref_P WFVM}.
    Definition eval_Soundness_Ref := ifold_ WFVM _ (ip_algebra (iPAlgebra := eval_Soundness_Ref_WFVM)).

    (* DeRef case. *)

    Definition eval_Soundness_DeRef_P (i : WFValueM_i D V MT ME _) :=
      WFValueM _ _ _ _ _ WFVM i /\
      WFValueMC _ _ _ _ _ WFVM (wfvm_S _ _ _ _ _ i)
      (do v <- wfvm_v _ _ _ _ _ i;
        match isLoc (proj1_sig v) with
          | Some n =>
            do env <- get;
              return_
              match lookup env n with
                | Some mv => mv
                | None => stuck' 457
              end
          | None => return_ (stuck' 456)
        end)
      (do te <- wfvm_T _ _ _ _ _ i;
        match isTRef (proj1_sig te) with
          | Some t => return_ t
          | None => fail
        end).

    Inductive eval_Soundness_DeRef_Name := eval_soundness_deref_name.

    Global Instance WFVM_State_eval_Soundness_DeRef :
       iPAlgebra eval_Soundness_DeRef_Name eval_Soundness_DeRef_P (WFValueM_State TypContext).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
        try eassumption; unfold eval_Soundness_DeRef_P; split; simpl; intros.
      (* WFVM_Get *)
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      intros env WF_env; apply (proj1 (H0 env WF_env)).
      rewrite <- associativity; apply (inject_i (subGF := Sub_WFVM_State_WFVM));
        econstructor; intros; eapply H0; eauto.
      (* WFVM_Put *)
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      apply (proj1 H2).
      unfold wbind; rewrite <- associativity.
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      apply (proj2 H2).
    Qed.

    Context {WF_invertLoc'_WFV :
      iPAlgebra WF_invertLoc'_Name (WF_invertLoc'_P _ WFV) WFV}.
    Context {WF_invertLoc_WFV :
      iPAlgebra WF_invertLoc_Name (WF_invertLoc_P _ WFV) WFV}.

    Global Instance wfvm_Base_eval_Soundness_DeRef :
      iPAlgebra eval_Soundness_DeRef_Name eval_Soundness_DeRef_P (WFValueM_base D V MT ME _ WFV).
    Proof.
      constructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (WFV := WFV)
        (Monad_ME := Mon_M) (Fail_MT := Fail_MT);
        try eassumption; unfold eval_Soundness_DeRef_P; simpl; split; intros.
      (* WFVM_Return *)
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor; eauto.
      repeat rewrite <- left_unit.
      unfold isLoc.
      caseEq (project (G := LocValue) (proj1_sig v)).
      destruct l.
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; intros.
      destruct H1 as [mt' mt'_eq]; rewrite <- mt'_eq.
      apply project_inject in H2.
      generalize (WF_invertLoc' TypContext WFV _ _ H0 _ H2) as WF_v_T; intros;
        inversion WF_v_T; subst.
      rewrite H7 in H2; unfold loc, inject in H2; simpl in H2;
        apply in_t_UP'_inject in H2; repeat rewrite wf_functor in H2;
          simpl in H2; apply (f_equal (prj (sub_F := LocValue))) in H2;
            repeat rewrite prj_inj in H2; injection H2; intros; subst.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor 1 with (T := T'); intros.
      rename H3 into WF_env_Sigma.
      destruct (WF_EnvLookup _ _ _ _ WF_env_Sigma H6) as [v' [lookup_v' WF_v'_T']];
        unfold Value in *; rewrite lookup_v'; auto.
      apply bind_orn; intros; destruct a as [T'' T''_eq]; subst; simpl.
      rewrite <- T''_eq, H8.
      unfold isTRef, tref, project; simpl; rewrite out_in_fmap, fmap_fusion;
        rewrite wf_functor; simpl; rewrite prj_inj; auto with typeclass_instances.
      apply return_orn; rewrite in_out_UP'_inverse; auto.
      exact (proj2_sig _).
      auto with typeclass_instances.
      exact (proj2_sig _).
      destruct H1 as [mt' mt'_eq].
      rewrite <- mt'_eq.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM));
        apply WFVM_Untyped'_T; intros.
      rewrite <- H1.
      unfold isTRef; caseEq (project (G := RefType) (proj1_sig T)); auto.
      destruct r; apply project_inject in H3; generalize (WF_invertLoc _ _ _ _ H0 _ (proj2_sig s));
        eauto with typeclass_instances; try (intros; exact (proj2_sig _)).
      revert H3; unfold inject, tref; simpl; repeat rewrite wf_functor; simpl;
        intros; apply H4 in H3; inversion H3; subst;
          rewrite H9 in H2; rewrite project_loc_loc in H2; discriminate; auto.
      (* WFVM_Untyped' *)
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor; eauto.
      unfold wbind; rewrite <- associativity, bind_fail;
        apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor 2.
    Qed.

    Context {eval_Soundness_DeRef_WFVM : iPAlgebra eval_Soundness_DeRef_Name eval_Soundness_DeRef_P WFVM}.
    Definition eval_Soundness_DeRef := ifold_ WFVM _ (ip_algebra (iPAlgebra := eval_Soundness_DeRef_WFVM)).

    (* Assign Case *)

    Context {eq_DType_eq_DT : PAlgebra eq_DType_eqName D D (UP'_P (eq_DType_eq_P D))}.
    Context {eq_DType_neq_D : PAlgebra eq_DType_neqName D D (UP'_P (eq_DType_neq_P D))}.

    Definition eval_Soundness_Assign_e2_P (i : WFValueM_i D V MT ME TypContext) :=
      WFValueM _ _ _ _ _ WFVM i /\
      forall n (T T' : DType) (mt : MT {T'0 : Names.DType D | proj1_sig T = proj1_sig T'0}),
        SigLookup (wfvm_S _ _ _ _ _ i) n = Some T' ->
        WFValueMC _ _ _ _ _ WFVM (wfvm_S _ _ _ _ _ i)
        (do v2 <- (wfvm_v _ _ _ _ _ i);
          do env <- get; put (replace_el n v2 env) >> return_ unit')
        (do t1 <- mt; do t2 <- (wfvm_T _ _ _ _ _ i);
          (if eq_DType D (proj1_sig T') t2 then return_ tunit' else fail)).

    Inductive eval_Soundness_Assign_e2_Name := eval_soundness_assign_e2_name.

    Global Instance WFVM_State_eval_Soundness_Assign_e2 :
       iPAlgebra eval_Soundness_Assign_e2_Name eval_Soundness_Assign_e2_P (WFValueM_State TypContext).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
        try eassumption; unfold eval_Soundness_Assign_e2_P; split; simpl; intros.
      (* WFVM_Get *)
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      intros env WF_env; apply (proj1 (H0 env WF_env)).
      rewrite <- associativity; apply (inject_i (subGF := Sub_WFVM_State_WFVM));
        econstructor; intros; eapply H0; eauto.
      (* WFVM_Put *)
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      apply (proj1 H2).
      unfold wbind; rewrite <- associativity.
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      eapply H2; eauto.
      eapply ConsExtension_SigLookup; eauto.
    Qed.

    Global Instance wfvm_Base'_eval_Soundness_Assign_e2 :
      iPAlgebra eval_Soundness_Assign_e2_Name eval_Soundness_Assign_e2_P (WFValueM_base D V MT ME _ WFV).
    Proof.
      constructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (WFV := WFV)
        (Monad_ME := Mon_M) (Fail_MT := Fail_MT);
        try eassumption; unfold eval_Soundness_Assign_e2_P; split; simpl; intros.
      (* WFVM_Return *)
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor; eauto.
      destruct H1 as [mt' mt'_eq].
      repeat rewrite <- left_unit.
      caseEq (eq_DType D  (proj1_sig T) T').
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; intros.
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor 2 with (Sigma' := Sigma);
        intros; auto.
      eapply WF_EnvReplace; eauto.
      destruct T' as [T' T'_UP']; apply (WFV_proj1_b D V _ WFV _ _ H0 T' T'_UP');
          simpl in *.
      apply sym_eq; apply (eq_DType_eq D T _ H1).
      apply ConsExtension_id.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor.
      apply (inject_i (subGF := Sub_WFV_Unit_WFV)); econstructor.
      unfold unit; simpl; auto.
      instantiate (1 := tunit'); reflexivity.
      apply bind_orn'; intros.
      rewrite <- mt'_eq; eapply bind_orn; intros.
      destruct a as [T2 T2_eq]; destruct a0 as [T3 T3_eq];
        subst; simpl.
      caseEq (eq_DType D (proj1_sig T') T3).
      apply return_orn; simpl; auto.
      exists fail; rewrite fmap_fail; auto.
      rewrite associativity.
      rewrite <- mt'_eq, fmap_m, associativity, <- fmap_m.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)).
      apply WFVM_Untyped'_T; intros T2; intros.
      caseEq (eq_DType D (proj1_sig T') T2); auto.
      rewrite <- (eq_DType_eq D _ _ H4) in H3; rewrite H3 in H1.
      elimtype False; apply (eq_DType_neq D _ _ H1); auto.
      (* WFVM_Untyped *)
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor; eauto.
      unfold wbind; rewrite <- associativity; rewrite bind_fail.
      repeat rewrite associativity.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor.
    Qed.

    Context {eval_Soundness_Assign_e2_WFVM :
      iPAlgebra eval_Soundness_Assign_e2_Name eval_Soundness_Assign_e2_P WFVM}.
    Definition eval_Soundness_Assign_e2 :=
      ifold_ WFVM _ (ip_algebra (iPAlgebra := eval_Soundness_Assign_e2_WFVM)).

    Definition eval_Soundness_Assign_e1_P (i : WFValueM_i D V MT ME _) :=
      WFValueM _ _ _ _ _ WFVM i /\
      forall mE mT,
        (forall Sigma', ConsExtension Sigma' (wfvm_S _ _ _ _ _ i) -> 
          WFValueMC _ _ _ _ _ WFVM Sigma' mE mT) ->
        WFValueMC _ _ _ _ _ WFVM (wfvm_S _ _ _ _ _ i) 
        (do v <- wfvm_v _ _ _ _ _ i;
          match isLoc (proj1_sig v) with
            | Some n0 =>
              do v2 <- mE;
                do env <- get; put (replace_el n0 v2 env) >> return_ unit'
            | None => return_ (stuck' 3)
          end)
        (do t1 <- wfvm_T _ _ _ _ _ i;
          match isTRef (proj1_sig t1) with
            | Some t =>
              do t2 <- mT;
                (if eq_DType D (proj1_sig t) t2 then return_ tunit' else fail)
            | None => fail
          end).

    Inductive eval_Soundness_Assign_e1_Name := eval_soundness_assign_e1_name.

    Global Instance WFVM_State_eval_Soundness_Assign_e1 :
       iPAlgebra eval_Soundness_Assign_e1_Name eval_Soundness_Assign_e1_P (WFValueM_State TypContext).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
        try eassumption; unfold eval_Soundness_Assign_e1_P; split; simpl; intros.
      (* WFVM_Get *)
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      intros env WF_env; apply (proj1 (H0 env WF_env)).
      rewrite <- associativity; apply (inject_i (subGF := Sub_WFVM_State_WFVM));
        econstructor; intros; eapply H0; eauto.
      (* WFVM_Put *)
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      apply (proj1 H2).
      unfold wbind; rewrite <- associativity.
      apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
      apply (proj2 H2); auto.
      intros; eapply H3; eauto.
      eapply ConsExtension_trans; eauto.
    Qed.

    Global Instance wfvm_Base_eval_Soundness_Assign_e1 :
      iPAlgebra eval_Soundness_Assign_e1_Name eval_Soundness_Assign_e1_P (WFValueM_base D V MT ME _ WFV).
    Proof.
      constructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (WFV := WFV)
        (Monad_ME := Mon_M) (Fail_MT := Fail_MT);
        try eassumption; unfold eval_Soundness_Assign_e1_P; split; simpl; intros.
      (* WFVM_Return *)
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor; eauto.
      repeat rewrite <- left_unit.
      unfold isLoc; caseEq (project (G := LocValue) (proj1_sig v)).
      destruct l.
      destruct H1 as [mt' mt'_eq].
      apply project_inject in H3.
      generalize (WF_invertLoc' _ _ _ _ H0 _ H3) as WF_v_T;
        intros; inversion WF_v_T; subst.
      assert (do t1 <- fmap (@proj1_sig _ _) mt';
      match isTRef (proj1_sig t1) with
      | Some t =>
          do t2 <- mT;
          (if eq_DType D (proj1_sig t) t2 then return_ tunit' else fail)
      | None => fail
      end = do t1 <- mt';
            do t2 <- mT;
              (if eq_DType D (proj1_sig T') t2 then return_ tunit' else fail)).
      rewrite <- eq_bind_fmap with
        (g := @proj1_sig _ _) (ma := mt')
        (f := fun t1 : {T' : Names.DType D | proj1_sig T = proj1_sig T'} =>
          do t2 <- mT;
            (if eq_DType D (proj1_sig T') t2 then return_ tunit' else fail)).
      apply f_equal; apply functional_extensionality; intros.
      apply f_equal; apply functional_extensionality; intros; auto.
      destruct a as [T'' T''_eq]; simpl.
      rewrite <- T''_eq.
      rewrite H8; unfold isTRef, tref, project; simpl; rewrite wf_functor; simpl;
        rewrite out_in_fmap, wf_functor; simpl; rewrite prj_inj; simpl.
      generalize in_out_UP'_inverse; simpl; intros H10; rewrite H10; auto.
      exact (proj2_sig _).
      rewrite H1.
      apply (eval_Soundness_Assign_e2 _ (H2 Sigma (ConsExtension_id _))); auto.
      rewrite H7 in H3; unfold loc, inject in H3; simpl in H3;
        apply in_t_UP'_inject in H3; repeat rewrite wf_functor in H3;
          simpl in H2; apply (f_equal (prj (sub_F := LocValue))) in H3;
            repeat rewrite prj_inj in H3; injection H3; intros; subst; auto.
      auto with typeclass_instances.
      exact (proj2_sig _).
      assert (do t1 <- mt;
      match isTRef (proj1_sig t1) with
      | Some t =>
          do t2 <- mT;
            (if eq_DType D (proj1_sig t) t2 then return_ tunit' else fail)
      | None => fail
      end = do t1 <- mt; fail).
      destruct H1 as [mt' mt'_eq].
      rewrite <- mt'_eq.
      rewrite <- eq_bind_fmap with
        (g := @proj1_sig _ _) (ma := mt')
        (f := fun t1 => match isTRef (proj1_sig (proj1_sig t1)) with
                          | Some t =>
                            do t2 <- mT;
                              (if eq_DType D (proj1_sig t) t2 then return_ tunit' else fail)
                          | None => fail
                        end); auto.
      rewrite <- eq_bind_fmap with
        (g := @proj1_sig _ _) (ma := mt')
        (f := fun _ => fail); auto.
      apply f_equal; apply functional_extensionality; intros;
        destruct x as [T' T'_eq]; simpl.
      unfold isTRef; caseEq (project (G := RefType) (proj1_sig T')); auto.
      destruct r; apply project_inject in H1; generalize (WF_invertLoc _ _ _ _ H0 _ (proj2_sig s));
        eauto with typeclass_instances; try (intros; exact (proj2_sig _)).
      revert H1; unfold inject, tref; simpl; repeat rewrite wf_functor; simpl;
        intros; rewrite T'_eq in H4; apply H4 in H1; inversion H1; subst;
          rewrite H9 in H3; rewrite project_loc_loc in H3; discriminate; auto.
      rewrite H4.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor 2.
      (* WFVM_Untyped *)
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor; eauto.
      unfold wbind; rewrite <- associativity, bind_fail.
      apply (inject_i (subGF := Sub_WFVM_Base_WFVM)); econstructor 2.
    Qed.

    Context {eval_Soundness_Assign_e1_WFVM :
      iPAlgebra eval_Soundness_Assign_e1_Name eval_Soundness_Assign_e1_P WFVM}.
    Definition eval_Soundness_Assign_e1 :=
      ifold_ WFVM _ (ip_algebra (iPAlgebra := eval_Soundness_Assign_e1_WFVM)).

    Lemma eval_soundness_H (typeof_R eval_R : Set) 
      typeof_rec eval_rec
      : 
      forall (eT : typeof_R) (eE : eval_R) Sigma
        (IHe : WFValueMC _ _ _ _ _ WFVM Sigma
          (eval_rec eE) (typeof_rec eT)),
        WFValueMC _ _ _ _ _ WFVM Sigma
        (RefE_evalM _ eval_rec (Ref _ eE))
        (RefE_typeof _ typeof_rec (Ref _ eT)).
    Proof.
      simpl; intros.
      apply (eval_Soundness_Ref _ IHe).
    Qed.

    Lemma eval_soundness_H0 (typeof_R eval_R : Set) 
      typeof_rec eval_rec
      : 
      forall (eT : typeof_R) (eE : eval_R) Sigma        
        (IHe : WFValueMC _ _ _ _ _ WFVM Sigma 
          (eval_rec eE) (typeof_rec eT)),
        WFValueMC _ _ _ _ _ WFVM Sigma
        (RefE_evalM _ eval_rec (DeRef _ eE))
        (RefE_typeof _ typeof_rec (DeRef _ eT)).
    Proof.
      simpl; intros.
      apply (eval_Soundness_DeRef _ IHe).
    Qed.

    Lemma eval_soundness_H1 (typeof_R eval_R : Set) 
      typeof_rec eval_rec
      : 
      forall (e1T e2T : typeof_R) (e1E e2E : eval_R) Sigma
        (IHe1 : WFValueMC _ _ _ _ _ WFVM Sigma 
          (eval_rec e1E) (typeof_rec e1T))
        (IHe2 : forall Sigma', ConsExtension Sigma' Sigma -> 
          WFValueMC _ _ _ _ _ WFVM Sigma' 
          (eval_rec e2E) (typeof_rec e2T)),
        WFValueMC _ _ _ _ _ WFVM Sigma
        (RefE_evalM _ eval_rec (Assign _ e1E e2E))
        (RefE_typeof _ typeof_rec (Assign _ e1T e2T)).
    Proof.
      simpl; intros.
      apply (eval_Soundness_Assign_e1 _ (IHe1)); eauto.
    Qed.

    Section PSoundness.
      
      Variable P_bind : Set.
      Variable E' : Set -> Set.
      Context {Fun_E' : Functor E'}.
      Context {Sub_RefE_E' : RefE :<: E'}.
      Context {WF_Fun_E' : WF_Functor _ _ Sub_RefE_E' _ _}.

      Variable P : UP'_F E -> UP'_F E' -> P_bind -> TypContext -> Prop.

      Variable Ref_Decomp : 
        forall e1 e2 ref1_UP' ref2_UP' pb Sigma, 
          P (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Ref _ e2)))) ref2_UP')
          (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Ref _ e1)))) ref1_UP')
          pb Sigma -> 
          forall e2_UP' e1_UP',
            P (exist _ (proj1_sig (P := Universal_Property'_fold) e2) e2_UP') 
            (exist _ (proj1_sig (P := Universal_Property'_fold) e1) e1_UP') pb Sigma.

      Variable DeRef_Decomp : 
        forall e1 e2 ref1_UP' ref2_UP' pb Sigma, 
          P (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (DeRef _ e2)))) ref2_UP')
          (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (DeRef _ e1)))) ref1_UP')
          pb Sigma -> 
          forall e2_UP' e1_UP',
            P (exist _ (proj1_sig (P := Universal_Property'_fold) e2) e2_UP') 
            (exist _ (proj1_sig (P := Universal_Property'_fold) e1) e1_UP') pb Sigma.

      Variable Assign_Decomp1 : 
        forall e11 e12 e21 e22 assign1_UP' assign2_UP' pb Sigma, 
          P (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Assign _ e12 e22)))) assign2_UP')
          (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Assign _ e11 e21)))) assign1_UP')
          pb Sigma -> 
          forall e12_UP' e11_UP',
            P (exist _ (proj1_sig (P := Universal_Property'_fold) e12) e12_UP') 
            (exist _ (proj1_sig (P := Universal_Property'_fold) e11) e11_UP') pb Sigma.

      Variable Assign_Decomp2 : 
        forall e11 e12 e21 e22 assign1_UP' assign2_UP' pb Sigma, 
          P (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Assign _ e12 e22)))) assign2_UP')
          (exist Universal_Property'_fold
            (in_t (fmap (@proj1_sig _ _) (inj (Assign _ e11 e21)))) assign1_UP')
          pb Sigma -> 
          forall e22_UP' e21_UP',
            P (exist _ (proj1_sig (P := Universal_Property'_fold) e22) e22_UP') 
            (exist _ (proj1_sig (P := Universal_Property'_fold) e21) e21_UP') pb Sigma.

      Global Instance Ref_eval_soundness'
      (P_Sub : forall e e' pb Sigma Sigma', 
        P e e' pb Sigma -> ConsExtension Sigma' Sigma -> 
        P e e' pb Sigma')
      {Typeof_E' : forall T, FAlgebra TypeofName T (typeofR D MT) E'}
      {WF_typeof_E' : forall T, @WF_FAlgebra TypeofName T _ _ _
        Sub_RefE_E' (MAlgebra_typeof_RefE T) (Typeof_E' _)}
      (pb : P_bind)
      (eval_rec : Exp -> evalMR V ME)
      (typeof_rec : UP'_F E' -> typeofR D MT)
      :
      P2Algebra ES'_ExpName RefE E' E 
      (UP'_P2 (eval_soundness'_P D V E MT ME _ WFVM
        _ _ _ P pb typeof_rec eval_rec (f_algebra (FAlgebra := Typeof_E' _ ))
        (f_algebra (FAlgebra := evalM_E)))).
    Proof.
      apply ind_P2Alg_RefE; eauto.
      (* ref case *)
      destruct e as [eE eT]; destruct IHe as [[UP'_eE UP'_eT] IHe].
      assert (Universal_Property'_fold
        (inject
          (Ref (sig Universal_Property'_fold)
            (exist Universal_Property'_fold eE (proj1 (conj UP'_eE UP'_eT))))) /\
        Universal_Property'_fold
        (inject
          (Ref (sig Universal_Property'_fold)
            (exist Universal_Property'_fold eT (proj2 (conj UP'_eE UP'_eT)))))) as WF_Ref 
      by (simpl; split; unfold inject; exact (proj2_sig _)).
      unfold eval_soundness'_P, UP'_P2; simpl; intros;
        constructor 1 with (x := WF_Ref).
      unfold inject; simpl;
        repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      intros eval_rec_proj typeof_rec_proj Sigma WF_Sigma IHa.
      repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold RefE_fmap.
      apply eval_soundness_H.
      generalize (IHa _ _ (exist _ _ UP'_eE, exist _ _ UP'_eT) 
        (Ref_Decomp _ _ _ _ _ _ WF_Sigma _ _)); 
        simpl.
      generalize (typeof_rec_proj (exist _ _ UP'_eE)); simpl; 
        intros eE_eq WF_ref; rewrite <- eE_eq; eauto.
      apply WF_ref.
      apply IHe; eauto.
      apply (Ref_Decomp _ _ _ _ _ _ WF_Sigma _ _).
      (* deref case *)
      destruct e as [eE eT]; destruct IHe as [[UP'_eE UP'_eT] IHe].
      assert (Universal_Property'_fold
        (inject
          (DeRef (sig Universal_Property'_fold)
            (exist Universal_Property'_fold eE (proj1 (conj UP'_eE UP'_eT))))) /\
        Universal_Property'_fold
        (inject
          (DeRef (sig Universal_Property'_fold)
            (exist Universal_Property'_fold eT (proj2 (conj UP'_eE UP'_eT)))))) as WF_DeRef 
      by (simpl; split; unfold inject; exact (proj2_sig _)).
      unfold eval_soundness'_P, UP'_P2; simpl; intros;
        constructor 1 with (x := WF_DeRef).
      unfold inject; simpl;
        repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      intros eval_rec_proj typeof_rec_proj Sigma WF_Sigma IHa.
      repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold RefE_fmap.
      apply eval_soundness_H0.
      generalize (IHa _ _ (exist _ _ UP'_eE, exist _ _ UP'_eT)
      (DeRef_Decomp _ _ _ _ _ _ WF_Sigma _ _)); simpl.
      generalize (typeof_rec_proj (exist _ _ UP'_eE)); simpl; 
        intros eE_eq WF_ref; rewrite <- eE_eq; eauto.
      apply WF_ref.
      apply IHe; eauto.
      apply (DeRef_Decomp _ _ _ _ _ _ WF_Sigma _ _).
      (* assign case *)
      destruct e1 as [e1E e1T]; destruct e2 as [e2E e2R];
        destruct IHe1 as [[UP'_e1E UP'_e1T] IHe1];
          destruct IHe2 as [[UP'_e2E UP'_e2T] IHe2]; simpl in *|-*.
      unfold eval_soundness'_P, UP'_P2; intros.
      assert (Universal_Property'_fold
        (fst
          (inject
            (Assign (sig Universal_Property'_fold)
              (exist Universal_Property'_fold e1E
                (proj1 (conj UP'_e1E UP'_e1T)))
              (exist Universal_Property'_fold e2E
                (proj1 (conj UP'_e2E UP'_e2T)))),
            inject
            (Assign (sig Universal_Property'_fold)
              (exist Universal_Property'_fold e1T
                (proj2 (conj UP'_e1E UP'_e1T)))
              (exist Universal_Property'_fold e2R
                (proj2 (conj UP'_e2E UP'_e2T)))))) /\
        Universal_Property'_fold
        (snd
          (inject
            (Assign (sig Universal_Property'_fold)
              (exist Universal_Property'_fold e1E
                (proj1 (conj UP'_e1E UP'_e1T)))
              (exist Universal_Property'_fold e2E
                (proj1 (conj UP'_e2E UP'_e2T)))),
            inject
            (Assign (sig Universal_Property'_fold)
              (exist Universal_Property'_fold e1T
                (proj2 (conj UP'_e1E UP'_e1T)))
              (exist Universal_Property'_fold e2R
                (proj2 (conj UP'_e2E UP'_e2T))))))) as WF_Assign 
      by (simpl; split; unfold inject; exact (proj2_sig _)).
      constructor 1 with (x := WF_Assign); simpl.
      unfold inject; simpl;
        repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      intros eval_rec_proj typeof_rec_proj Sigma WF_Sigma IHa.
      repeat rewrite out_in_fmap; repeat rewrite wf_functor; simpl.
      rewrite (wf_algebra (WF_FAlgebra := WF_eval_F));
        rewrite (wf_algebra (WF_FAlgebra := WF_typeof_E' _));
          simpl fmap; simpl f_algebra; unfold RefE_fmap.
      intros; apply eval_soundness_H1.
      generalize (IHa _ _ (exist _ _ UP'_e1E, exist _ _ UP'_e1T) 
        (Assign_Decomp1 _ _ _ _ _ _ _ _ WF_Sigma _ _)); simpl.
      generalize (typeof_rec_proj (exist _ _ UP'_e1E)); simpl; 
        intros e1E_eq WF_ref; rewrite <- e1E_eq; eauto.
      apply WF_ref.
      apply IHe1; eauto.
      apply (Assign_Decomp1 _ _ _ _ _ _ _ _ WF_Sigma _ _).
      intros Sigma' ConsExt_Sig'_Sig;
        generalize (IHa _ _ (exist _ _ UP'_e2E, exist _ _ UP'_e2T)
          (Assign_Decomp2 _ _ _ _ _ _ _ _ 
            (P_Sub _ _ _ _ _ WF_Sigma ConsExt_Sig'_Sig) _ _)); simpl.
      generalize (typeof_rec_proj (exist _ _ UP'_e2E)); simpl; 
        intros e2E_eq WF_ref; rewrite <- e2E_eq; eauto.
      apply WF_ref.
      apply IHe2; eauto.
      apply (Assign_Decomp2 _ _ _ _ _ _ _ _
            (P_Sub _ _ _ _ _ WF_Sigma ConsExt_Sig'_Sig) _ _).
    Defined.

  End PSoundness.

  Global Instance Ref_eval_soundness'' typeof_rec eval_rec :
    P2Algebra ES'_ExpName RefE E E
    (UP'_P2 (eval_soundness'_P D V E MT ME _ WFVM
      _ _ _ (fun _ _ _ _ => True) tt typeof_rec eval_rec
      (f_algebra (FAlgebra := Typeof_F _)) (f_algebra (FAlgebra := evalM_E)))).
  Proof.
    eapply Ref_eval_soundness'; eauto.
  Defined.
  
  Section State_Sound_Sec.
    
    Variable WFV' : (WFValue_i D V (list DType) -> Prop) -> WFValue_i D V (list DType) -> Prop.
    Variable funWFV' : iFunctor WFV'.
    
    Variable WFVM' : (WFValueM_i D V MT ME (list DType) -> Prop) -> WFValueM_i D V MT ME (list DType) -> Prop.
    Variable funWFVM' : iFunctor WFVM'.

    Global Instance DType_Env_CE : ConsExtensionC (list DType) :=
      {| ConsExtension := fun Sigma' Sigma => 
        forall n T, lookup Sigma n = Some T -> lookup Sigma' n = Some T |}.
    Proof.
      (* ConsExtension_id *)
      eauto.
      (* ConsExtension_trans *)
      eauto.
    Defined.
      
    Global Instance DType_Env_S : SigTypContextC (list DType) :=
      {| SigLookup := lookup;
        SigInsert := insert _|}.

    Global Instance DType_Env_WFE : WF_EnvC (list DType) :=
      {| WF_Env := fun env Sigma =>
        P2_Env (WFValueC _ _ _ WFV' Sigma) env Sigma |}.

    Global Instance DType_Env_SCE : SigConsExtensionC (list DType) _ _.
    Proof. 
      constructor.
    (* ConsExtension_SigLookup *) 
      eauto.
    (* ConsExtension_SigInsert *)
      simpl; intros; erewrite lookup_Some_insert; eauto.
    Qed.

    Context {WFV_Weaken_WFV' : iPAlgebra WFValue_Weaken_Name (WFValue_Weaken_P D V _ WFV') WFV'}.

    Definition WFV'_Weaken := ifold_ WFV' _ (ip_algebra (iPAlgebra := WFV_Weaken_WFV')).

    Global Instance DType_Env_SWFE : Sig_WF_EnvC (list DType) WFV' _ _.
    Proof.
      constructor; simpl; intros.
      (* WF_EnvLookup *)
      unfold WF_Env in H; simpl in H.
      apply (P2_Env_lookup' _ _ _ _ _ H); auto.
      (* WF_EnvInsertLookup *)
      unfold Value.
      rewrite (P2_Env_length _ _ _ _ _ H); apply lookup_insert.
      (* WF_EnvInsert *)
      apply P2_Env_insert; eauto.
      eapply P2_Env_P_P'; eauto.
      intros; generalize (WFV'_Weaken _ H1).
      unfold WFValue_Weaken_P; simpl; intros; eapply H2.
      simpl; intros; erewrite lookup_Some_insert; eauto.
      (* WF_EnvReplace *)
      eapply P2_Env_replace_el; eauto.
    Qed.

    Definition State_Sound_P (i : WFValueM_i D V MT ME (list DType)) :=
      forall T (env : list Value),
        WF_Env env (wfvm_S _ _ _ _ _ i) ->
        wfvm_T _ _ _ _ _ i = return_ T ->
        exists v : Value, exists env', exists Sigma',
          (put env) >> wfvm_v _ _ _ _ _ i = put env' >> return_ (M := ME) v /\
          WFValueC _ _ _ WFV' Sigma' v T.

    Inductive State_Sound_Name := State_Sound_name.

    Context {Reasonable_MT : Reasonable_Monad MT}.

    Context {WFV_proj1_a_WFV' : iPAlgebra WFV_proj1_a_Name (WFV_proj1_a_P D V _ WFV') WFV'}.
    Context {WFV_proj1_b_WFV' : iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV') WFV'}.

    Global Instance State_Sound_WFVM_State :
      iPAlgebra State_Sound_Name State_Sound_P (WFValueM_State (list DType)).
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_State with (TypContextCE := DType_Env_CE)
        (TypContext_WFE := DType_Env_WFE);
        try assumption; unfold State_Sound_P; simpl; intros.
      (* WFVM_Get *)
      destruct (H0 env H1 T0 env H1 H2) as [v [env' [Sigma' [eval_eq WF_v_T]]]].
      exists v; exists env'; exists Sigma'.
      unfold wbind; rewrite associativity.
      generalize put_get as put_get'; intros; unfold wbind in put_get'; rewrite put_get'.
      rewrite <- associativity.
      rewrite <- left_unit.
      split; unfold wbind; auto.
      (* WFVM_Put *)
      unfold wbind; rewrite associativity.
      generalize put_put as put_put'; intros; unfold wbind in put_put'; rewrite put_put'.
      destruct (H2 T0 env H0 H4) as [v [env' [Sigma'' [eval_eq WF_v_T]]]].
      exists v; exists env'; exists Sigma''; split; auto.
    Qed.
    
    Global Instance State_Sound_WFVM_base :
      iPAlgebra State_Sound_Name State_Sound_P (WFValueM_base D V MT ME _ WFV').
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (Monad_ME := Mon_M)
        (Fail_MT := Fail_MT) (WFV := WFV');
        try assumption; unfold State_Sound_P; simpl; intros.
      exists v; exists env; exists Sigma; split; auto.
      destruct H1 as [mt' mt'_eq]; subst.
      destruct (fmap_exists _ _ _ _ _ H3) as [[T' T_eq] T'_eq].
      simpl in *; subst; auto.
      destruct T0 as [T0 T0_UP'].
      apply (WFV_proj1_b _ _ _ WFV' _ _ H0); simpl; auto.
      (* WFVM_Untyped' *)
      simpl in H1; apply sym_eq in H1.
      apply FailMonad_Disc in H1; destruct H1; auto.
    Qed.

    Context {State_Sound_WFVM : iPAlgebra State_Sound_Name State_Sound_P WFVM'}.

    Context {eval_soundness'_Exp_E : forall (typeof_rec : UP'_F E -> typeofR D MT)
       (eval_rec : Names.Exp E -> evalMR V ME),
       P2Algebra ES'_ExpName E E E 
       (UP'_P2
         (eval_soundness'_P D V E MT ME _ WFVM'
           Datatypes.unit E Fun_E
           (fun _ _ _ _ => True)
           tt typeof_rec eval_rec f_algebra
           (f_algebra (FAlgebra := evalM_E' (@Names.Exp E Fun_E)))))}.

     Context {WF_MAlg_typeof : WF_MAlgebra Typeof_F}.
     Context {WF_MAlg_eval : WF_MAlgebra evalM_E'}.

    Lemma eval_State_soundness : forall (e : Exp) Sigma,
      WFValueMC _ _ _ _ _ WFVM' Sigma (evalM (evalM_E := evalM_E') _ _ _ (proj1_sig e)) 
      (typeof _ _ MT (proj1_sig e)).
    Proof.
      intro; rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig e) (proj2_sig e)).
      simpl; unfold typeof, evalM, fold_, mfold, in_t.
      repeat rewrite wf_malgebra; unfold mfold.
      destruct (Ind2 (Ind_Alg := eval_soundness'_Exp_E
        (fun e => typeof _ _ _ (proj1_sig e)) 
        (fun e => evalM (evalM_E := evalM_E') _ _ _ (proj1_sig e)))
        _ (proj2_sig e)) as [e' eval_e'].
      unfold eval_soundness'_P in eval_e'.
      intros; eapply eval_e'; intros; auto; try constructor.
      rewrite (@in_out_UP'_inverse _ _ (proj1_sig _) (proj2_sig _)); auto.
      rewrite (@in_out_UP'_inverse _ _ (proj1_sig _) (proj2_sig _)); auto.
      rewrite <- (@in_out_UP'_inverse _ _ (proj1_sig (fst a)) (proj2_sig _)).
      unfold typeof, evalM, mfold; simpl; unfold in_t; 
        repeat rewrite wf_malgebra; unfold mfold; apply H; auto.
    Qed.

    Theorem eval_State_Sound :
      forall (Sigma : Env DType) (e : Exp) (T : DType) (env : Env Value),
        WF_Env env Sigma ->
        typeof D E MT (proj1_sig e) = return_ T ->
        exists v : Value,
          exists env' : Env Value,
            exists Sigma' : Env DType,
              (put env) >> evalM (evalM_E := evalM_E') _ _ _ (proj1_sig e) = 
              put env' >> return_ (M := ME) v /\
              WFValueC _ _ _ WFV' Sigma' v T.
    Proof.
      intros Sigma e.
      apply (ifold_ WFVM' _ (ip_algebra (iPAlgebra := State_Sound_WFVM)) _
        (eval_State_soundness e Sigma)).
    Qed.

  End State_Sound_Sec.

End Ref.

(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
