Require Import FJ_tactics.
Require Import List.
Require Import Functors.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import MonadLib.
Require Import Ref.
Require Import Arith.

Section Ref_Arith.

  Variable F : Set -> Set.
  Context {Fun_F : Functor F}.
  Context {Sub_Except_F : RefE :<: F}.

  Context {WF_Sub_Except_F : WF_Functor _ _ Sub_Except_F _ _}.
  
  Variable V : Set -> Set.
  Context {Fun_V : Functor V}.
  
  Context {Sub_StuckValue_V : StuckValue :<: V}.
  
  Context {Sub_BotValue_V : BotValue :<: V}.
  Context {Sub_LocValue_V : LocValue :<: V}.
  Context {Sub_NatValue_V : NatValue :<: V}.
  Context {Sub_UnitValue_V : UnitValue :<: V}.
  
  Variable (ME : Set -> Set). (* Evaluation Monad. *)
  Context {Fun_M : Functor ME}.
  Context {Mon_M : Monad ME}.
  Context {StateM_ME : StateM ME (list (Names.Value V))}.

  Variable D : Set -> Set.
  Context {Fun_D : Functor D}.
  Context {Sub_RefType_D  : RefType :<: D}.
  Context {Sub_AType_D : AType :<: D}.
  Context {Sub_UnitType_D : UnitType :<: D}.

  Context {WF_Sub_UnitType_D : WF_Functor _ _ Sub_UnitType_D _ _}.
  Context {WF_Sub_AType_D : WF_Functor _ _ Sub_AType_D _ _}.
  Context {WF_Sub_RefType_D : WF_Functor _ _ Sub_RefType_D _ _}.

  Variable MT : Set -> Set.
  Context {Fun_MT : Functor MT}.
  Context {Mon_MT : Monad MT}.
  Context {Fail_MT : FailMonad MT}.

  Variable TypContext : Set.

  Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
  Context {funWFV : iFunctor WFV}.

  Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> WFValueM_i D V MT ME TypContext -> Prop.
  Variable funWFVM : iFunctor WFVM.

  Context {Typeof_F : forall T, FAlgebra TypeofName T (typeofR D MT) F}.
  Context {evalM_F : forall T, FAlgebra EvalName T (evalMR V ME) F}.

  Context {TypContextCE : ConsExtensionC TypContext}.
  Context {TypContext_WFE : WF_EnvC V TypContext}.

  Context {WFV_Weaken_WFV : iPAlgebra WFValue_Weaken_Name (WFValue_Weaken_P D V _ WFV) WFV}.

  Context {Sub_WFVM_State_WFVM : Sub_iFunctor (WFValueM_State D V MT ME TypContext) WFVM}.

  Global Instance WFVM_State'_eval_Soundness_Addm :
    iPAlgebra Arith_eval_Soundness_Addm_Name
    (Arith_eval_Soundness_Addm_P V ME D MT TypContext WFVM) (WFValueM_State D V MT ME TypContext).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
      try eassumption; unfold Arith_eval_Soundness_Addm_P; split; simpl; intros.
    (* WFVM_Get' *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor.
    intros env WF_env; apply (H0 env WF_env).
    rewrite <- associativity; apply (inject_i (subGF := Sub_WFVM_State_WFVM)).
    econstructor; intros; eapply H0; eauto.
    (* WFVM_Put' *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto; apply (proj1 H2).
    unfold wbind; rewrite <- associativity; apply (inject_i (subGF := Sub_WFVM_State_WFVM)).
    econstructor; intros; generalize (proj2 H2); simpl; intros; eauto.
    eapply H4; eauto.
    intros; eapply H3; eapply ConsExtension_trans; eauto.
  Qed.

  Global Instance WFVM_State_eval_Soundness_Addn :
    iPAlgebra Arith_eval_Soundness_Addn_Name
    (Arith_eval_Soundness_Addn_P V ME D MT TypContext WFV WFVM) (WFValueM_State D V MT ME TypContext).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_State;
      try eassumption; unfold Arith_eval_Soundness_Addn_P; split; simpl; intros.
    (* WFVM_Get *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor.
    intros env WF_env; apply (H0 env WF_env).
    rewrite <- associativity; apply (inject_i (subGF := Sub_WFVM_State_WFVM));
      econstructor; intros; eapply H0; eauto.
    (* WFVM_Put *)
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto; apply (proj1 H2).
    unfold wbind; rewrite <- associativity.
    apply (inject_i (subGF := Sub_WFVM_State_WFVM)); econstructor; eauto.
    apply H2; eauto.
    eapply (WFV_Weaken D V _ WFV _ _ WF_v_T); simpl; eauto.
  Qed.

  Context {Dis_TUnit_TNat : Distinct_Sub_Functor _ Sub_UnitType_D Sub_AType_D}.
  Context {Dis_TRef_TNat : Distinct_Sub_Functor _ Sub_RefType_D Sub_AType_D}.

  Global Instance WF_invertVI_Unit :
    iPAlgebra WF_invertVI_Name (WF_invertVI_P V D _ WFV) (WFValue_Unit D V _).
  Proof.
    econstructor.
    unfold iAlgebra; unfold WF_invertVI_P; intros; inversion H; simpl in *.
    destruct i; injection H3; intros; subst; simpl in H0.
    rewrite H0 in H2.
    elimtype False.
    unfold tnat, tunit, tnat', tunit' in H2.
    generalize (inject_discriminate Dis_TUnit_TNat
      (TUnit (sig Universal_Property'_fold)) (TNat (sig Universal_Property'_fold))).
    revert H2; unfold inject, inject'; eauto.
  Qed.

  Context {TypContext_S : SigTypContextC D TypContext}.

  Global Instance WF_invertVI_Loc :
    iPAlgebra WF_invertVI_Name (WF_invertVI_P V D _ WFV) (WFValue_Loc D V _).
  Proof.
    econstructor.
    unfold iAlgebra; unfold WF_invertVI_P; intros; inversion H; simpl in *.
    destruct i; injection H4; intros; subst; simpl in H0.
    rewrite H0 in H3.
    elimtype False.
    unfold tnat, tref, tnat', tref' in H3.
    generalize (inject_discriminate Dis_TRef_TNat
      (TRef (sig Universal_Property'_fold) T') (TNat (sig Universal_Property'_fold))).
    revert H3; unfold inject, inject'; destruct T'; eauto.
  Qed.

  Global Instance WF_invertLoc_VI :
   iPAlgebra WF_invertLoc_Name (WF_invertLoc_P D V _ WFV) (WFValue_VI V D _).
  Proof.
    econstructor.
    unfold iAlgebra; unfold WF_invertLoc_P; intros; inversion H; simpl in *.
    destruct i; injection H3; intros; subst; simpl in H0.
    rewrite H0 in H2.
    elimtype False.
    unfold tnat, tref, tnat', tref' in H2.
    generalize (inject_discriminate Dis_TRef_TNat
      (TRef (sig Universal_Property'_fold) (exist _ T' T'_UP)) (TNat (sig Universal_Property'_fold))).
    revert H3; unfold inject, inject'; eauto.
  Qed.

  Context {Dis_LocValue_NatValue : Distinct_Sub_Functor _ Sub_LocValue_V Sub_NatValue_V}.
  Context {WF_Sub_LocValue_V : WF_Functor _ _ Sub_LocValue_V _ _}.
  Context {WF_Sub_NatValue_V : WF_Functor _ _ Sub_NatValue_V _ _}.

  Global Instance WF_invertLoc'_VI :
   iPAlgebra WF_invertLoc'_Name (WF_invertLoc'_P D V _ WFV) (WFValue_VI V D _).
  Proof.
    econstructor.
    unfold iAlgebra; unfold WF_invertLoc'_P; intros; inversion H; simpl in *.
    destruct i; injection H3; intros; subst; simpl in H0.
    rewrite H0 in H1.
    elimtype False.
    unfold loc, vi, loc', vi' in H1.
    generalize (inject_discriminate Dis_LocValue_NatValue (Loc _ n) (VI _ n0)).
    revert H1; unfold inject, inject'; eauto.
  Qed.

  End Ref_Arith.

(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
