Require Import FJ_tactics.
Require Import List.
Require Import Functors.
Require Import Names.
Require Import FunctionalExtensionality.
Require Import MonadLib.
Require Import Exception.
Require Import Ref.

Section Ref_Exception.

  Variable D : Set -> Set.
  Context {Fun_D : Functor D}.
  Context {Sub_RefType_D  : RefType :<: D}.
  Context {Sub_UnitType_D : UnitType :<: D}.
  Context {WF_Sub_UnitType_D : WF_Functor _ _ Sub_UnitType_D _ _}.
  Context {WF_Sub_RefType_D : WF_Functor _ _ Sub_RefType_D _ _}.

  Variable E : Set -> Set.
  Context {Fun_E : Functor E}.
  Context {Sub_RefE_E : RefE :<: E}.
  Context {Sub_ExceptE_E : ExceptE D :<: E}.
  Context {WF_Sub_ExceptE_E : WF_Functor _ _ Sub_ExceptE_E _ _}.
  Context {WF_Sub_RefE_E : WF_Functor _ _ Sub_RefE_E _ _}.
  
  Variable V : Set -> Set.
  Context {Fun_V : Functor V}.
  Context {Sub_LocValue_V : LocValue :<: V}.
  Context {Sub_StuckValue_V : StuckValue :<: V}.
  Context {Sub_UnitValue_V : UnitValue :<: V}.

  Variable (ME : Set -> Set). (* Evaluation Monad. *)
  Context {Fun_M : Functor ME}.
  Context {Mon_M : Monad ME}.
  Context {Exception_ME : Exception ME Datatypes.unit}.
  Context {StateM_ME : StateM ME (list (Names.Value V))}.
  Context {Fail_ME : FailMonad ME}.

  Variable MT : Set -> Set.
  Context `{Fail_MT : FailMonad MT}.
  Context {Inj_MT : InjMonad MT}.
  Context {Reasonable_MT : Reasonable_Monad MT}.

  Context {eq_DType_DT : forall T, FAlgebra eq_DTypeName T (eq_DTypeR D) D}.
  Context {eq_DType_eq_DT : PAlgebra eq_DType_eqName D D (UP'_P (eq_DType_eq_P D))}.
  Context {eq_DType_neq_D : PAlgebra eq_DType_neqName D D (UP'_P (eq_DType_neq_P D))}.

  Context {Typeof_E : forall T, FAlgebra TypeofName T (typeofR D MT) E}.
  Context {WF_ExceptE_Typeof_E : forall T, @WF_FAlgebra TypeofName _ _ (ExceptE D) E
    Sub_ExceptE_E (MAlgebra_typeof_ExceptE _ MT _) (Typeof_E T)}.
  Context {WF_RefE_Typeof_E : forall T, @WF_FAlgebra TypeofName _ _ RefE E
    Sub_RefE_E (MAlgebra_typeof_RefE _ MT _) (Typeof_E T)}.

  Context {evalM_E : FAlgebra EvalName (Exp E) (evalMR V ME) E}.
  Context {evalM_E' : forall T, FAlgebra EvalName T (evalMR V ME) E}.
  Context {WF_ExceptE_eval_E : @WF_FAlgebra EvalName _ _ (ExceptE D) E
    Sub_ExceptE_E (MAlgebra_evalM_ExceptE D V ME _) evalM_E}.
  Context {WF_RefE_eval_E : @WF_FAlgebra EvalName _ _ RefE E
    Sub_RefE_E (MAlgebra_evalM_RefE V ME _) evalM_E}.

  Variable TypContext : Set.
  Variable WFV : (WFValue_i D V TypContext -> Prop) -> WFValue_i D V TypContext -> Prop.
  Context {funWFV : iFunctor WFV}.

  Variable WFVM : (WFValueM_i D V MT ME TypContext -> Prop) -> WFValueM_i D V MT ME TypContext -> Prop.
  Variable funWFVM : iFunctor WFVM.

  Context {TypContextCE : ConsExtensionC TypContext}.
  Context {TypContext_WFE : WF_EnvC V TypContext}.

  Context {Sub_WFVM_Base_WFVM : Sub_iFunctor (WFValueM_base D V MT ME TypContext WFV) WFVM}.
  Context {Sub_WFVM_State_WFVM : Sub_iFunctor (WFValueM_State D V MT ME TypContext) WFVM}.
  Context {Sub_WFVM_Except_WFVM : Sub_iFunctor (WFValueM_Except D V MT ME _) WFVM}.
  
  Global Instance WFVM_Except_eval_Soundness_Ref :
    iPAlgebra eval_Soundness_Ref_Name (eval_Soundness_Ref_P D V MT ME _ WFVM) (WFValueM_Except D V MT ME _).
  Proof.
    constructor.
    unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
      try eassumption; unfold eval_Soundness_Ref_P; split; simpl; intros.
    (* WFVM_Throw' *)
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
    simpl; rewrite bind_throw; apply (inject_i (subGF := Sub_WFVM_Except_WFVM));
      constructor.
    (* WFVM_Catch' *)
    destruct H0 as [H0 H0'].
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
    intros t Sigma' Sub_Sig_Sig'; apply (proj1 (H1 t Sigma' Sub_Sig_Sig')).
    repeat rewrite <- associativity.
    rewrite catch_msimp.
    apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2.
    assert (forall T,
      fmap (@proj1_sig _ _) (do U <- kT'; do c <- kT U T; return_ (tref' D c)) =
      do c <- (fmap (@proj1_sig _ _) (kT'>>= fun U => kT U T));
        return_ (in_t (@inj (RefType) D _ _ (TRef _ c)))) by
    (intros; rewrite fmap_m, associativity, fmap_m;
      repeat rewrite <- associativity; apply f_equal;
        apply functional_extensionality; intros; apply f_equal;
          apply functional_extensionality; intros;
            repeat rewrite <- left_unit; simpl; rewrite wf_functor;
              reflexivity).
    intros; repeat rewrite H2; erewrite kt_eq; eauto.
    rewrite associativity, <- catch_msimp, associativity; apply H0.
    intros; rewrite associativity, <- catch_msimp, associativity; apply H1;
      eauto.
  Qed.

    Global Instance WFVM_Except_eval_Soundness_DeRef :
       iPAlgebra eval_Soundness_DeRef_Name (eval_Soundness_DeRef_P D V MT ME _ WFVM) (WFValueM_Except D V MT ME _).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
        try eassumption; unfold eval_Soundness_DeRef_P; split; simpl; intros.
      (* WFVM_Throw' *)
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
      simpl; rewrite bind_throw; apply (inject_i (subGF := Sub_WFVM_Except_WFVM));
        constructor.
      (* WFVM_Catch' *)
      destruct H0 as [H0 H0'].
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
      intros t Sigma' Sub_Sig_Sig'; apply (proj1 (H1 t Sigma' Sub_Sig_Sig')).
      repeat rewrite <- associativity.
      rewrite catch_msimp.
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2.
      assert (forall T,
        fmap (@proj1_sig _ _)
        (do U <- kT';
          do c <- kT U T;
            match isTRef D (proj1_sig c) with
              | Some t => return_ t
              | None => fail
            end) = fmap (@proj1_sig _ _)
        (do c <- (fmap (@proj1_sig _ _) (kT' >>= fun U => kT U T));
          match isTRef D c with
            | Some t => return_ t
            | None => fail
          end)) as H4 by
      (intros; rewrite fmap_m, associativity, fmap_m, <- associativity, fmap_m;
        repeat rewrite <- associativity;
          apply f_equal;
            apply functional_extensionality; intros; apply f_equal;
              apply functional_extensionality; intros;
                repeat rewrite <- left_unit; reflexivity).
      intros; repeat rewrite H4; erewrite kt_eq; eauto.
      rewrite associativity, <- catch_msimp, associativity; apply H0.
      intros; rewrite associativity, <- catch_msimp, associativity; apply H1;
        eauto.
    Qed.

    Context {TypContext_S : SigTypContextC D TypContext}.
    Context {TypContext_SCE : SigConsExtensionC D TypContext _ _}.

    Global Instance WFVM_State_eval_Soundness_Assign_e2 :
       iPAlgebra eval_Soundness_Assign_e2_Name (eval_Soundness_Assign_e2_P D V MT ME _ WFVM) (WFValueM_Except D V MT ME _).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
        try eassumption; unfold eval_Soundness_Assign_e2_P; split; simpl; intros.
      (* WFVM_Throw' *)
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
      simpl; rewrite bind_throw; apply (inject_i (subGF := Sub_WFVM_Except_WFVM));
        constructor.
      (* WFVM_Catch' *)
      destruct H0 as [H0 H0'].
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
      intros t Sigma' Sub_Sig_Sig'; apply (proj1 (H1 t Sigma' Sub_Sig_Sig')).
      repeat rewrite <- associativity.
      rewrite curry_msimp with (mta := mt).
      rewrite catch_msimp.
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2.
      assert (forall T,
        fmap (@proj1_sig _ _)
        (do U <- do a <- mt; do b <- kT'; return_ (a, b);
          do c <- kT (snd U) T;
            (if eq_DType D (proj1_sig T') c then return_ (tunit' D) else fail)) =
        do a <- mt; (do c <- (fmap (@proj1_sig _ _) (kT' >>= fun U => kT U T));
          (if eq_DType D c T' then return_ (tunit D) else fail))).
      (intros; rewrite fmap_m, associativity, fmap_m;
        repeat rewrite <- associativity; apply f_equal;
            apply functional_extensionality; intros; rewrite <- associativity;
              apply f_equal;
                apply functional_extensionality; intros; rewrite <- left_unit;
                  simpl; apply f_equal; apply functional_extensionality; intros;
                    rewrite <- left_unit).
      caseEq (eq_DType D (proj1_sig T') x1);
      caseEq (eq_DType D (proj1_sig x1) T').
      rewrite <- left_unit; auto.
      elimtype False; apply (eq_DType_neq _ _ _ H4); generalize (eq_DType_eq _ _ _ H3); auto.
      elimtype False; apply (eq_DType_neq _ _ _ H3); generalize (eq_DType_eq _ _ _ H4); auto.
      rewrite bind_fail; auto.
      intros; repeat rewrite H3; erewrite kt_eq; eauto.
      rewrite associativity, <- catch_msimp, associativity.
      repeat rewrite <- associativity.
      rewrite rewrite_do with (n := mt)
        (m' := (fun _ : {T'0 : Names.DType D | proj1_sig T = proj1_sig T'0} =>
          do t2 <- do U <- kT'; mte >>= kT U;
            (if eq_DType D (proj1_sig T') t2 then return_ (tunit' D) else fail))).
      rewrite associativity; apply H0; auto.
      apply functional_extensionality; intros; repeat rewrite <- associativity;
        apply f_equal; apply functional_extensionality; intros; rewrite <- left_unit;
          repeat rewrite <- associativity; apply f_equal; apply functional_extensionality;
            intros; simpl; auto.
      intros; rewrite associativity, <- catch_msimp, associativity.
      repeat rewrite <- associativity.
      rewrite rewrite_do with (n := mt)
        (m' := (fun _ : {T'0 : Names.DType D | proj1_sig T = proj1_sig T'0} =>
          do t2 <- do U <- kT'; mth >>= kT U;
            (if eq_DType D (proj1_sig T') t2 then return_ (tunit' D) else fail))).
      rewrite associativity; apply H1; auto.
      eapply ConsExtension_SigLookup; eauto.
      apply functional_extensionality; intros; repeat rewrite <- associativity;
        apply f_equal; apply functional_extensionality; intros; rewrite <- left_unit;
          repeat rewrite <- associativity; apply f_equal; apply functional_extensionality;
            intros; simpl; auto.
    Qed.

    Global Instance WFVM_State_eval_Soundness_Assign_e1 :
       iPAlgebra eval_Soundness_Assign_e1_Name (eval_Soundness_Assign_e1_P D V MT ME _ WFVM) (WFValueM_Except D V MT ME _).
    Proof.
      constructor.
      unfold iAlgebra; intros; eapply ind_alg_WFVM_Except;
        try eassumption; unfold eval_Soundness_Assign_e1_P; split; simpl; intros.
      (* WFVM_Throw' *)
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); constructor.
      simpl; rewrite bind_throw; apply (inject_i (subGF := Sub_WFVM_Except_WFVM));
        constructor.
      (* WFVM_Catch' *)
      destruct H0 as [H0 H0'].
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2; eauto.
      intros t Sigma' Sub_Sig_Sig'; apply (proj1 (H1 t Sigma' Sub_Sig_Sig')).
      repeat rewrite <- associativity.
      rewrite catch_msimp.
      apply (inject_i (subGF := Sub_WFVM_Except_WFVM)); econstructor 2.
      intros; apply f_equal.
      generalize (fun T' => kt_eq_msimp _ kT' (fun a => kT a T')
        (fun c => match isTRef D c with
          | Some t =>
            do t2 <- mT;
              (if eq_DType D (proj1_sig t) t2 then return_ (tunit' D) else fail)
          | None => fail
        end) (@proj1_sig _ _)); unfold DType; intro Heq; repeat rewrite Heq.
      erewrite kt_eq; eauto.
      rewrite associativity, <- catch_msimp, associativity; eapply H0; eauto.
      intros; rewrite associativity, <- catch_msimp, associativity; eapply H1;
        eauto.
      simpl; intros; eapply H2; eapply ConsExtension_trans; eauto.
    Qed.

  Section Except_State_Sound_Sec.

    Variable WFV' : (WFValue_i D V (list (DType D)) -> Prop) -> WFValue_i D V (list (DType D)) -> Prop.
    Variable funWFV' : iFunctor WFV'.
    
    Variable WFVM' : (WFValueM_i D V MT ME (list (DType D)) -> Prop) -> 
      WFValueM_i D V MT ME (list (DType D)) -> Prop.
    Variable funWFVM' : iFunctor WFVM'.

    Definition Except_State_Sound_P (i : WFValueM_i D V MT ME (list (DType D))) :=
      forall T (env : list (Value V)),
        WF_Env (WF_EnvC := DType_Env_WFE D V WFV') _ env (wfvm_S _ _ _ _ _ i) ->
        fmap (@proj1_sig _ _) (wfvm_T _ _ _ _ _ i) = return_ (proj1_sig T) ->
        (exists v : Value V, exists env', exists Sigma' : list (DType D),
          (put env) >> wfvm_v _ _ _ _ _ i = put env' >> return_ (M := ME) v /\
          WFValueC D V _ WFV' Sigma' v T) \/
        (exists t, exists env', exists Sigma' : list (DType D),
          put env >> wfvm_v _ _ _ _ _ i = put env' >> throw t
        /\ WF_Env (WF_EnvC := DType_Env_WFE D V WFV') _ env' Sigma'
        /\ (forall n T, lookup (wfvm_S _ _ _ _ _ i) n = Some T -> lookup Sigma' n = Some T)).

    Inductive Except_State_Sound_Name := except_state_sound_name.

    Global Instance Except_State_Sound_WFVM_State :
      iPAlgebra Except_State_Sound_Name Except_State_Sound_P 
      (WFValueM_State (TypContext_WFE := DType_Env_WFE _ _ WFV') D V MT ME (list (DType D))).
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_State with (State_M := StateM_ME)
        (TypContextCE := DType_Env_CE _) (TypContext_WFE := DType_Env_WFE _ _ WFV');
        try assumption; unfold Except_State_Sound_P; simpl; intros.
      (* WFVM_Get *)
      destruct (H0 env H1 T0 env H1 H2) as [[v [env' [Sigma' [eval_eq WF_v_T]]]]
        | [[] [env' [Sigma' [k_eq [WF_env_Sigma' Sigma'_Cons]]]]]].
      left; exists v; exists env'; exists Sigma'.
      unfold wbind; rewrite associativity.
      generalize put_get as put_get'; intros; unfold wbind in put_get'; rewrite put_get'.
      rewrite <- associativity.
      rewrite <- left_unit.
      split; unfold wbind; auto.
      right; exists tt; exists env'; exists Sigma';
        unfold wbind; rewrite associativity; repeat split; auto.
      generalize put_get; unfold wbind; intros pg; rewrite pg;
        rewrite <- associativity; rewrite <- left_unit; auto.
      (* WFVM_Put *)
      unfold wbind; rewrite associativity.
      generalize put_put as put_put'; intros; unfold wbind in put_put'; rewrite put_put';
        clear put_put'.
      destruct (H2 T0 env H0 H4) as [[v [env' [Sigma'' [eval_eq WF_v_T]]]]
        | [[] [env' [Sigma'' [k_eq [WF_env_Sigma'' Sigma''_Cons]]]]]].
      left; exists v; exists env'; exists Sigma''; split; auto.
      right; exists tt; exists env'; exists Sigma''; split; auto.
    Qed.

    Context {WFV_proj1_b_WFV : iPAlgebra WFV_proj1_b_Name (WFV_proj1_b_P D V _ WFV') WFV'}.

    Global Instance Except_State_Sound_WFVM_Base :
      iPAlgebra Except_State_Sound_Name Except_State_Sound_P (WFValueM_base D V MT ME _ WFV').
    Proof.
      econstructor.
      unfold iAlgebra; intros; apply ind_alg_WFVM_base with (WFV := WFV')
        (Fail_MT := Fail_MT) (Monad_ME := Mon_M);
        try assumption; unfold Except_State_Sound_P; simpl; intros.
      (* WFVM_Return' *)
      left; exists v; exists env; exists Sigma; split; auto.
      destruct H1 as [mt' mt'_eq]; subst.
      rewrite fmap_fusion in H3.
      destruct (fmap_exists _ _ _ _ _ H3) as [[T' T_eq] T'_eq].
      simpl in *; subst; auto.
      destruct T0 as [T0 T0_UP'].
      apply (WFV_proj1_b _ _ _ _ _ _ H0); simpl; auto.
      simpl in T'_eq; congruence.
      (* WFVM_Untyped' *)
      simpl in H1; apply sym_eq in H1.
      unfold wbind in H1;
        rewrite fmap_m, <- associativity, bind_fail in H1.
      apply FailMonad_Disc in H1; destruct H1; auto.
    Defined.

    Context {ME_eq_dec' : forall (A : Set) (mte : ME A) (env : list (Names.Value (Fun_V := Fun_V) V)),
      (exists a : A, exists env' : list (Names.Value (Fun_V := Fun_V) V),
        wbind (H := Mon_M) (put (StateM := StateM_ME) env) 
        mte = wbind (H := Mon_M) (put (StateM := StateM_ME) env') (return_ (Monad := Mon_M) a)) \/
      (exists env', put env >> mte = put env' >> throw tt)}.
    Context {MT_eq_dec : forall (A : Set) (mta : MT A),
      {exists a, mta = return_ a} + {mta = fail}}.

    Section Except_State_Sound_WFVM_Except_Sec.
      (* Section with one possible put_catch law. *)
      Context {put_catch : forall (A : Set) (env : list (Value V)) e h,
        put env >>= (fun _ => catch (A := A) e h) = catch (put env >>= fun _ => e) h}.
      Context {put_throw : forall (A B : Set) (env env' : list (Value V)) t,
        put env >>= (fun _ => throw t (A := A)) = put env' >>= (fun _ => throw t) ->
        put env >>= (fun _ => throw t (A := B)) = put env' >>= (fun _ => throw t)}.
      Context {Put_Exception_Disc :
        forall (A : Set) (a : A) env env' n,
          (put env >>= fun _ => return_ a) <> put env' >>= fun _ => throw n}.

      Global Instance Except_State_Sound_WFVM_Except :
        iPAlgebra Except_State_Sound_Name Except_State_Sound_P (WFValueM_Except D V MT ME _).
      Proof.
        econstructor.
        unfold iAlgebra; intros; apply ind_alg_WFVM_Except with (Exception_ME := Exception_ME)
          (Fail_MT := Fail_MT) (eq_DType_DT := eq_DType_DT) (TypContextCE := DType_Env_CE _);
          try assumption; unfold Except_State_Sound_P; simpl; intros.
      (* throw case *)
        simpl; right; exists tt; exists env; exists Sigma; split; auto.
      (* catch case *)
        destruct (MT_eq_dec _ mte) as [[T' mte_eq] | mte_eq]; subst.
        destruct (MT_eq_dec _ mth) as [[T'' mth_eq] | mth_eq]; subst.
        repeat rewrite <- left_unit in H3.
        caseEq (eq_DType _ (proj1_sig T') T''); rewrite H4 in H3.
        destruct (H0 _ _ H2 H3) as [[v' [env' [Sigma' [ek'_eq WF_v'_T']]]] |
          [[] [env' [Sigma' [ek'_eq [WF_env'_Sigma' Sigma'_Cons]]]]]].
        destruct (ME_eq_dec' _ e' env) as [[v'' [env'' e'_eq]] | [env'' e'_eq]];
          subst.
        left; exists v'; exists env'; exists Sigma'; split; auto.
        unfold wbind; rewrite associativity, put_catch.
        unfold Value in *|-*; unfold wbind in e'_eq, ek'_eq; rewrite e'_eq, <- put_catch,
          catch_return, <- ek'_eq, associativity, e'_eq; auto.
      (* unfold Value in *|-*; unfold wbind in *|-*; rewrite associativity in ek'_eq;
         rewrite e'_eq in ek'_eq; rewrite <- associativity, bind_fail in ek'_eq.
         elimtype False; eapply Put_FailMonad_Disc with (a := v') (mb := put env'') (env := env');
         eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto. *)
        unfold Value in *|-*; unfold wbind in *|-*; rewrite associativity in ek'_eq;
          rewrite e'_eq in ek'_eq; rewrite <- associativity, bind_throw in ek'_eq.
        elimtype False; eapply Put_Exception_Disc with (a := v') (env' := env'') (env := env');
          eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; eauto.
        destruct (ME_eq_dec' _ e' env) as [[v'' [env'' e'_eq]] | [env'' e'_eq]];
          subst.
        right; exists tt; exists env'; exists Sigma'; unfold wbind in *|-*; split; auto.
        unfold Value in *|-*; rewrite associativity, e'_eq in ek'_eq.
        rewrite associativity, put_catch, e'_eq, <- put_catch, catch_return; auto.
      (* unfold Value in *|-*; unfold wbind in *|-*; rewrite associativity in ek'_eq;
         rewrite e'_eq in ek'_eq; rewrite <- associativity, bind_fail in ek'_eq.
         elimtype False; eapply fail_neq_throw; eauto.*)
        unfold Value in *|-*; unfold wbind in *|-*; rewrite associativity, e'_eq, <- associativity in ek'_eq.
        rewrite rewrite_do with (m' := fun _ => throw tt) in ek'_eq.
        destruct (H1 tt Sigma' Sigma'_Cons T env' WF_env'_Sigma') as
          [[v'' [env''' [Sigma''' [ek''_eq WF_v''_T'']]]] |
            [[] [env''' [Sigma''' [ek''_eq [WF_env'''_Sigma''' Sigma'''_Cons]]]]]].
        rewrite <- H3.
        rewrite (rewrite_do) with (m' := fun U => kT U T'').
        rewrite (rewrite_do) with (m := fun U => return_ T' >>= _) (m' := fun U => kT U T').
        apply kt_eq; rewrite (eq_DType_eq _ _ _ H4); auto.
        apply functional_extensionality; intros; rewrite <- left_unit; auto.
        apply functional_extensionality; intros; rewrite <- left_unit; auto.
        left; exists v''; exists env'''; exists Sigma'''; split; auto.
        unfold wbind in *|-*; rewrite associativity, put_catch, e'_eq.
        rewrite (put_throw _ _ _ _ _ ek'_eq), <-put_catch, catch_throw',
          <- associativity; auto.
        right; exists tt; exists env'''; exists Sigma'''; repeat split; auto.
        unfold wbind in *|-*.
        unfold wbind in *|-*; rewrite associativity, put_catch, e'_eq.
        rewrite (put_throw _ _ _ _ _ ek'_eq), <-put_catch, catch_throw',
          <- associativity; auto.
        rewrite bind_throw; reflexivity.
        elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := kT');
          eauto with typeclass_instances; unfold wbind;
            rewrite fmap_m in H3; rewrite <- associativity in H3;
              rewrite rewrite_do with (m' := fun _ : C => fail) in H3; auto;
                apply functional_extensionality; intros; repeat rewrite bind_fail; auto.
        rewrite <- left_unit in H3.
        elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := kT');
          eauto with typeclass_instances; unfold wbind;
            rewrite fmap_m in H3; rewrite <- associativity in H3;
              rewrite rewrite_do with (m' := fun _ : C => fail) in H3; auto;
                apply functional_extensionality; intros; repeat rewrite bind_fail; auto.
        elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := kT');
          eauto with typeclass_instances; unfold wbind;
            rewrite fmap_m in H3; rewrite <- associativity in H3;
              rewrite rewrite_do with (m' := fun _ : C => fail) in H3; auto;
                apply functional_extensionality; intros; repeat rewrite bind_fail; auto.
      Qed.

    End Except_State_Sound_WFVM_Except_Sec.

    Section Except_State_Sound_WFVM_Except_Sec2.
      (* Section with the other possible put_catch law. *)
      Context {put_catch' : forall (A : Set) (env : list (Value V)) e h,
        put env >>= (fun _ => catch (A := A) e h) = 
        catch (put env >>= fun _ => e) (fun t => put env >>= fun _ => (h t))}.
      Context {Put_Exception_Disc :
        forall (A B : Set) (a : A) env env' n,
          (put env >>= fun _ => return_ a) <> put env' >>= fun _ => throw n}.

      Lemma put_catch'' : forall (A : Set) (env env' : list (Value V)) (e : ME A) h,
        catch (put env >>= fun _ => e) (fun t => put env' >>= fun _ => (h t)) =
        put env >> catch e (fun t => put env' >>= (fun _ => h t)).
      Proof.
        intros; generalize (put_put env env'); unfold wbind; intros put_put'.
        rewrite put_catch'.
        assert ((fun t : Datatypes.unit => put env' >>= (fun _ : Datatypes.unit => h t))
          = (fun t : Datatypes.unit =>
            put env >>=
            (fun _ : Datatypes.unit => put env' >>= (fun _ : Datatypes.unit => h t)))) as H
        by (apply functional_extensionality; destruct x; rewrite associativity, put_put'; auto).
        rewrite H; auto.
      Qed.

      Global Instance Except_State_Sound_WFVM_Except2 :
        iPAlgebra Except_State_Sound_Name Except_State_Sound_P (WFValueM_Except D V MT ME _).
      Proof.
        econstructor.
        unfold iAlgebra; intros; apply ind_alg_WFVM_Except with (Exception_ME := Exception_ME)
          (Fail_MT := Fail_MT) (eq_DType_DT := eq_DType_DT) (TypContextCE := DType_Env_CE _);
          try assumption; unfold Except_State_Sound_P; simpl; intros.
      (* throw case *)
        simpl; right; exists tt; exists env; exists Sigma; split; auto.
      (* catch case *)
        destruct (MT_eq_dec _ mte) as [[T' mte_eq] | mte_eq]; subst.
        destruct (MT_eq_dec _ mth) as [[T'' mth_eq] | mth_eq]; subst.
        repeat rewrite <- left_unit in H3.
        caseEq (eq_DType _ (proj1_sig T') T''); rewrite H4 in H3.
        destruct (H0 _ _ H2 H3) as [[v' [env' [Sigma' [ek'_eq WF_v'_T']]]] |
          [[] [env' [Sigma' [ek'_eq [WF_env'_Sigma' Sigma'_Cons]]]]]].
        destruct (ME_eq_dec' _ e' env) as [[v'' [env'' e'_eq]] | [env'' e'_eq]];
          subst.
        left; exists v'; exists env'; exists Sigma'; split; auto.
        unfold wbind; rewrite associativity, put_catch'.
        unfold Value in *|-*; unfold wbind in e'_eq, ek'_eq; rewrite e'_eq. 
        rewrite put_catch'', catch_return, <- ek'_eq, associativity, e'_eq; auto.
      (* unfold Value in *|-*; unfold wbind in *|-*; rewrite associativity in ek'_eq;
         rewrite e'_eq in ek'_eq; rewrite <- associativity, bind_fail in ek'_eq.
         elimtype False; eapply Put_FailMonad_Disc with (a := v') (mb := put env'') (env := env');
         eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; auto. *)
        unfold Value in *|-*; unfold wbind in *|-*; rewrite associativity in ek'_eq;
          rewrite e'_eq in ek'_eq; rewrite <- associativity, bind_throw in ek'_eq.
        elimtype False; eapply Put_Exception_Disc with (a := v') (env' := env'') (env := env');
          eauto with typeclass_instances; unfold wbind; rewrite <- left_unit; eauto.
        destruct (ME_eq_dec' _ e' env) as [[v'' [env'' e'_eq]] | [env'' e'_eq]];
          subst.
        right; exists tt; exists env'; exists Sigma'; unfold wbind in *|-*; split; auto.
        unfold Value in *|-*; rewrite associativity, e'_eq in ek'_eq.
        rewrite associativity, put_catch', e'_eq, put_catch'', catch_return; auto.
      (* unfold Value in *|-*; unfold wbind in *|-*; rewrite associativity in ek'_eq;
         rewrite e'_eq in ek'_eq; rewrite <- associativity, bind_fail in ek'_eq.
         elimtype False; eapply fail_neq_throw; eauto.*)
        unfold Value in *|-*; unfold wbind in *|-*; rewrite associativity, e'_eq, <- associativity in ek'_eq.
        rewrite rewrite_do with (m' := fun _ => throw tt) in ek'_eq.
        destruct (H1 tt Sigma (fun _ _ => id) T env H2) as
          [[v'' [env''' [Sigma''' [ek''_eq WF_v''_T'']]]] |
            [[] [env''' [Sigma''' [ek''_eq [WF_env'''_Sigma''' Sigma'''_Cons]]]]]].
        rewrite <- H3.
        rewrite (rewrite_do) with (m' := fun U => kT U T'').
        rewrite (rewrite_do) with (m := fun U => return_ T' >>= _) (m' := fun U => kT U T').
        apply kt_eq; rewrite (eq_DType_eq _ _ _ H4); auto.
        apply functional_extensionality; intros; rewrite <- left_unit; auto.
        apply functional_extensionality; intros; rewrite <- left_unit; auto.
        left; exists v''; exists env'''; exists Sigma'''; split; auto.
        generalize put_put; unfold wbind; intros put_put'.
        unfold wbind in *|-*; rewrite associativity, put_catch', e'_eq,
          put_catch'', catch_throw'; unfold wbind; rewrite associativity,
            put_put', <- associativity; auto.
        right; exists tt; exists env'''; exists Sigma'''; repeat split; auto.
        generalize put_put; unfold wbind in *|-*; intros put_put'.
        unfold wbind in *|-*; rewrite associativity, put_catch', e'_eq,
          put_catch'', catch_throw'; unfold wbind; rewrite associativity,
            put_put', <- associativity; auto.
        rewrite bind_throw; auto.
        elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := kT');
          eauto with typeclass_instances; unfold wbind;
            rewrite fmap_m in H3; rewrite <- associativity in H3;
              rewrite rewrite_do with (m' := fun _ : C => fail) in H3; auto;
                apply functional_extensionality; intros; repeat rewrite bind_fail; auto.
        rewrite <- left_unit in H3.
        elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := kT');
          eauto with typeclass_instances; unfold wbind;
            rewrite fmap_m in H3; rewrite <- associativity in H3;
              rewrite rewrite_do with (m' := fun _ : C => fail) in H3; auto;
                apply functional_extensionality; intros; repeat rewrite bind_fail; auto.
        elimtype False; eapply FailMonad_Disc with (M := MT) (a := (proj1_sig T)) (mb := kT');
          eauto with typeclass_instances; unfold wbind;
            rewrite fmap_m in H3; rewrite <- associativity in H3;
              rewrite rewrite_do with (m' := fun _ : C => fail) in H3; auto;
                apply functional_extensionality; intros; repeat rewrite bind_fail; auto.
      Qed.

    End Except_State_Sound_WFVM_Except_Sec2.
      
    Context {Except_State_Sound_WFVM : iPAlgebra Except_State_Sound_Name Except_State_Sound_P WFVM'}.
    Context {eval_soundness'_Exp_E : forall (typeof_rec : UP'_F E -> typeofR D MT)
      (eval_rec : Names.Exp E -> evalMR V ME),
      P2Algebra ES'_ExpName E E E
      (UP'_P2
        (eval_soundness'_P D V E MT ME _ WFVM'
          Datatypes.unit E Fun_E
          (fun _ _ _ _ => True) 
          tt typeof_rec eval_rec f_algebra 
          (f_algebra (FAlgebra := evalM_E' (@Names.Exp E Fun_E)))))}.
    Context {WF_MAlg_typeof : WF_MAlgebra Typeof_E}.
    Context {WF_MAlg_eval : WF_MAlgebra evalM_E'}.

    Theorem eval_Except_State_Sound :
      forall (e : Exp E) Sigma (T : DType D)
        (env : list (Value V)),
        WF_Environment D V _ WFV' Sigma env Sigma ->
        fmap (@proj1_sig _ _) (typeof D E MT (proj1_sig e)) = return_ (proj1_sig T) ->
        (exists v : Value V, exists env', exists Sigma',
          (put env) >> evalM (evalM_E := evalM_E') V E ME (proj1_sig e) = put env' >> return_ (M := ME) v /\
          WFValueC D V _ WFV' Sigma' v T) \/
        (exists t, exists env', exists Sigma',
          put env >> evalM (evalM_E := evalM_E') V E ME (proj1_sig e) = put env' >> throw t
        /\ WF_Environment D V _ WFV' Sigma' env' Sigma'
        /\ (forall n T, lookup Sigma n = Some T -> lookup Sigma' n = Some T)).
    Proof.
      intros e Sigma.
      apply (ifold_ WFVM' _ (ip_algebra (iPAlgebra := Except_State_Sound_WFVM)) _
        (eval_State_soundness (eval_soundness'_Exp_E := eval_soundness'_Exp_E) _ _ _ MT ME WFVM' e Sigma)).
    Qed.

  End Except_State_Sound_Sec.

  (* Context {SV : (SubValue_i V -> Prop) -> SubValue_i V -> Prop}.
  Context {Sub_SV_refl_SV : Sub_iFunctor (SubValue_refl V) SV}.
  Variable SVM : (SubValueM_i V ME -> Prop) -> SubValueM_i V ME -> Prop.
  Variable funSVM : iFunctor SVM.
  Context {Sub_SVM_Return_SVM : Sub_iFunctor (SubValueM_Return V ME SV) SVM}.
  Context {Sub_SVM_Bot_SVM : Sub_iFunctor (SubValueM_Bot V ME) SVM}.
  Context {Sub_SVM_Throw_SVM : Sub_iFunctor (SubValueM_Throw V ME) SVM}.
  Context {Sub_SVM_State_SVM : Sub_iFunctor (SubValueM_State V SV ME) SVM}.

  Class StateM_Exception := {
    get_catch : forall (A : Set) (k : list (Names.Value V) -> ME A) (h : nat -> ME A),
      catch (do env <- get; k env) h = do env <- get; catch (k env) h;
    put_catch : forall (A : Set) (env : list (Names.Value V))
      (k : Datatypes.unit -> ME A) (h : nat -> ME A),
      catch (do t <- put env; k t) h = do t <- put env; catch (k t) h}.

  Global Instance Parallel_bevalM_continuous_Catch {StateM_Exception_ME : StateM_Exception} :
    iPAlgebra bevalM_continuous_Catch_Name
    (bevalM_continuous_Catch_P F V ME SVM) (SubValueM_State V SV ME).
  Proof.
      apply ind_iPAlg_SVM_State; unfold bevalM_continuous_Catch_P; simpl; intros.
      Print bevalM_continuous_Catch_P.
      (* SVM_Get case *)
      repeat rewrite get_catch.
      apply (inject_i (subGF := Sub_SVM_State_SVM)); econstructor; simpl;
        intros; eapply H; eauto.
      (* SVM_Put case*)
      unfold wbind; repeat rewrite put_catch.
      apply (inject_i (subGF := Sub_SVM_State_SVM)); econstructor 2; simpl; eauto;
        intros; eapply H0; eauto.
    Defined.

    Global Instance Throw_bevalM_continuous_Ref  :
      iPAlgebra bevalM_continuous_Ref_Name (bevalM_continuous_Ref_P V ME SVM) (SubValueM_Throw V ME).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Ref_P;
        intros; inversion H; subst; simpl.
      rewrite bind_throw; apply (inject_i (subGF := Sub_SVM_Throw_SVM)); econstructor.
    Qed.

    Global Instance Throw_bevalM_continuous_DeRef  :
      iPAlgebra bevalM_continuous_DeRef_Name (bevalM_continuous_DeRef_P V ME SVM) (SubValueM_Throw V ME).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_DeRef_P;
        intros; inversion H; subst; simpl.
      rewrite bind_throw; apply (inject_i (subGF := Sub_SVM_Throw_SVM)); econstructor.
    Qed.

    Global Instance Throw_bevalM_continuous_Assign_Return  :
      iPAlgebra bevalM_continuous_Assign_Return_Name (bevalM_continuous_Assign_Return_P V ME SVM) (SubValueM_Throw V ME).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Assign_Return_P;
        intros; inversion H; subst; simpl.
      rewrite bind_throw; apply (inject_i (subGF := Sub_SVM_Throw_SVM)); econstructor.
    Qed.

    Global Instance Throw_bevalM_continuous_Assign  :
      iPAlgebra bevalM_continuous_Assign_Name (bevalM_continuous_Assign_P F V ME SVM) (SubValueM_Throw V ME).
    Proof.
      econstructor; unfold iAlgebra; intros; unfold bevalM_continuous_Assign_P;
        intros; inversion H; subst; simpl.
      repeat rewrite bind_throw; apply (inject_i (subGF := Sub_SVM_Throw_SVM)); econstructor.
    Qed. *)

End Ref_Exception.

(* Hint Extern 0 =>
  intros; match goal with
            |- (PAlgebra eval_Soundness_alg_Name
              (sig (UP'_P2 (eval_alg_Soundness_P _ _ _ _ _ _ _ _ _ _ _ _ _ _))) Arith) =>
            eapply Arith_eval_Soundness_alg; eauto with typeclass_instances
          end : typeclass_instances. *)


(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
