Require Import String.
Require Import Names.
Require Import Arith.
Require Import Functors.
Require Import MonadLib.
Require Import Ref.
Require Import Ref_Arith.

Open Scope string_scope.

Section Test_Section.

  Definition D := AType :+: UnitType :+: RefType.

  Definition E := Arith :+: RefE.

  Definition ex_1 : Names.Exp E :=
    (add' _ (lit' _ 1) (lit' _ 2)).
  Definition ex_2  : Names.Exp E :=
    (add' _ (lit' _ 5) (lit' _ 0)).
  Definition ex_3  : Names.Exp E :=
      (add' _ ex_1 ex_2).

  Definition V := StuckValue :+: NatValue :+: LocValue :+: UnitValue.
  Definition Functor_V : Functor V.
  Proof.
    eauto with typeclass_instances.
  Defined.

  Definition Functor_E : Functor E.
  Proof.
    eauto with typeclass_instances.
  Defined.

  Variable MT : Set -> Set.
  Context `{Fail_MT : FailMonad MT}.
  Context {Inj_MT : InjMonad MT}.
  Context {Reasonable_MT : Reasonable_Monad MT}.

  Variable ME : Set -> Set.
  Context `{State_ME : StateM ME (list (Value V))}.
  Context {Fail_ME : FailMonad ME}.

  Definition WFV := (WFValue_VI V D (list (Names.DType D))) ::+:: 
    (WFValue_Unit D V _) ::+:: (WFValue_Loc D V _).
  Definition WFVM := (WFValueM_base D V MT ME _ WFV) ::+:: 
    (WFValueM_State D V MT ME _
      (TypContextCE := DType_Env_CE _)
      (TypContext_WFE := DType_Env_WFE _ _ WFV)).

  Variable eq_DType_D :  PAlgebra eq_DType_eqName D D (UP'_P (eq_DType_eq_P D)).
  Variable neq_DType_D : PAlgebra eq_DType_neqName D D (UP'_P (eq_DType_neq_P D)).

  Instance typeof_alg : forall T, FAlgebra TypeofName T (typeofR D MT) (E).
  Proof.
    intros; eauto 150 with typeclass_instances.
  Defined.

  Lemma eval_Sound :
    forall (Sigma : Env (DType D)) (e : Exp E) (T : DType D) (env : Env (Value V)),
      WF_Environment D V _ WFV Sigma env Sigma ->
      typeof D E MT (proj1_sig e) = return_ T ->
      exists v : Value V,
        exists env' : Env (Value V),
          exists Sigma' : Env (DType D),
            (put env) >> evalM V (Fun_V := Functor_V) E ME (proj1_sig e) =
            put env' >> return_ (M := ME) v /\
            WFValueC D V _ WFV Sigma' v T.
  Proof.
    intros; eapply eval_State_Sound with (WFVM' := WFVM);
      eauto 10 with typeclass_instances.
    eauto 15 with typeclass_instances.
    intros; repeat apply @P2Algebra_Plus.
    eapply Arith_eval_soundness'' with (WFV := WFV);
      eauto 250 with typeclass_instances.
    eapply Ref_eval_soundness'';
      eauto 250 with typeclass_instances.
    repeat apply iPAlgebra_Plus;
      eauto  550 with typeclass_instances.
    apply wfvm_Base_eval_Soundness_Ref with (TypContextCE :=  _)
      (TypContext_S := _) (TypContext_WFE := DType_Env_WFE _ _ WFV);
      eauto  550 with typeclass_instances.
    repeat apply iPAlgebra_Plus;
      eauto  550 with typeclass_instances.
    apply wfvm_Base_eval_Soundness_DeRef with (TypContextCE := _)
      (TypContext_S := _) (TypContext_WFE := DType_Env_WFE _ _ WFV);
      eauto  550 with typeclass_instances.
    repeat apply iPAlgebra_Plus;
      eauto  550 with typeclass_instances.
    apply wfvm_Base_eval_Soundness_Assign_e1 with (TypContextCE := _)
      (TypContext_S := _);
      eauto 800 with typeclass_instances.
    repeat apply iPAlgebra_Plus;
      eauto  550 with typeclass_instances.
    apply wfvm_Base'_eval_Soundness_Assign_e2 with (TypContextCE := _)
      (TypContext_S := _) (TypContext_WFE := DType_Env_WFE _ _ WFV);
      eauto  550 with typeclass_instances.
    simpl; eauto.
  Qed.

  Eval compute in ("Soundness for 'Arith :+: References' Proven!").

End Test_Section.

(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
