Require Import String.
Require Import Names.
Require Import Lambda.
Require Import Functors.
Require Import PNames.
Require Import MonadLib.

Open Scope string_scope.

Section Test_Section.

  Definition D := LType.
  Definition E (A : Set) := Lambda D A.

  Global Instance Fun_E : forall (A : Set),
    Functor (E A).
  Proof.
    eauto with typeclass_instances.
  Defined.

  Definition V := StuckValue :+: (ClosValue E).
  Definition Functor_V : Functor V.
  Proof.
    eauto with typeclass_instances.
  Defined.

  Variable ME : Set -> Set.
  Context `{Mon_ME : Environment ME (list (Value V))}.
  Context {Fail_ME : FailMonad ME}.
  Context {FailEnvM : FailEnvironmentMonad (Env (Value V)) ME}.

  Definition Eqv := Lambda_eqv D E.
  Definition WFV := (WFValue_Clos D E V Eqv (list (Names.DType D))).
  Definition WFVM := (WFValueM_base D V MT ME _ WFV)
    ::+:: (WFValueM_Environment (TypContextCE := GammaTypContextCE _) D V ME _ WFV)
    ::+:: (WFValueM_Bot D V ME _).

  Instance eval_alg : FAlgebra EvalName (Exp E nat) (evalMR V ME) (E nat).
  Proof.
    apply (MAlgebra_evalM_Lambda D E V ME).
  Defined.

  Instance typeof_alg : forall T, FAlgebra TypeofName T (typeofR D MT) (E (DType D)).
  Proof.
    intros; eauto with typeclass_instances.
  Defined.

  Instance eqv_soundness_X_alg eval_rec :
    iPAlgebra soundness_X'_Name 
    (soundness_X'_P D V E ME _ Eqv WFVM (fun e => typeof _ _ MT (proj1_sig e)) eval_rec
      (f_algebra (FAlgebra := typeof_alg _))
      (f_algebra (FAlgebra := eval_alg))) (Eqv _ _).
  Proof.
    eapply eqv_soundness_X'_Lambda_eqv with (WFV := WFV)
      (TypContextCE := GammaTypContextCE _); eauto 250 with typeclass_instances .
  Qed.

  Instance Pure_Sound_X_WFVM : iPAlgebra Pure_Sound_X_Name (Pure_Sound_X_P D V ME WFV) WFVM.
  Proof.
    eauto 250 with typeclass_instances.
  Qed.

  Lemma eval_Sound :
      forall (n : nat) Sigma gamma gamma' gamma'' e' (e'' : UP'_F (E nat)),
        E_eqvC E Eqv gamma' gamma e'' e' ->
        forall (WF_gamma : forall n b, lookup gamma' n = Some b ->
          exists T, lookup gamma b = Some T)
        (WF_gamma'' : WF_Environment _ _ _ WFV Sigma gamma'' gamma)
        (WF_gamma2 : List.length gamma = List.length gamma')
        (WF_gamma' : forall n b, lookup gamma' n = Some b -> b = n)
        (WF_Sigma : Gamma _ Sigma = gamma) (T : DType D),
        typeof D (E (DType D)) MT (proj1_sig e') = return_ T ->
        local (M := ME) (fun _ => gamma'') (bevalM V E ME n e'') = (fail (A := Value V)) \/
        (exists v : Value V,
          local (M := ME) (fun _ => gamma'') (bevalM V E ME n e'') = return_ (M := ME) v /\
          WFValueC _ _ _ WFV Sigma v T).
  Proof.
    intros.
    apply eval_Pure_Sound_X with (EQV_E := Eqv) (WFVM := WFVM)
      (Typeof_E := typeof_alg) (gamma := gamma) (gamma' := gamma') (e' := e');
      eauto 250 with typeclass_instances.
    unfold WF_EnvG, DType in *|-*; rewrite WF_Sigma; auto.
  Qed.

  Eval compute in ("Type Soundness for Lambda Proven!").

End Test_Section.

(*
*** Local Variables: ***
*** coq-prog-args: ("-emacs-U" "-impredicative-set") ***
*** End: ***
*)
